#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import csv
import datetime
import os
import requests

import pandas as pd
import numpy as np

from glob import glob

from typing import Union, Iterable, Dict, Optional, Any

from .. import config, utils, core

# Set up logging
logger = config.logger


def from_swc(f: Union[str, pd.DataFrame, Iterable],
             connector_labels: Optional[Dict[str, Union[str, int]]] = {},
             soma_label: Union[str, int] = 1,
             include_subdirs: bool = False,
             **kwargs) -> 'core.NeuronObject':
    """Create Neuron/List from SWC file.

    This import is following format specified
    `here <http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html>`_

    Parameters
    ----------
    f :                 str | pandas.DataFrame | iterable
                        SWC string, URL, filename, folder or DataFrame.
                        If folder, will import all ``.swc`` files.
    connector_labels :  dict, optional
                        If provided will extract connectors from SWC.
                        Dictionary must map type to label:
                        ``{'presynapse': 7, 'postsynapse': 8}``
    include_subdirs :   bool, optional
                        If True and ``f`` is a folder, will also search
                        subdirectories for ``.swc`` files.

    **kwargs
                        Keyword arguments passed to ``navis.TreeNeuron``

    Returns
    -------
    navis.TreeNeuron
                        Contains SWC file header as ``.swc_header`` attribute.
    navis.NeuronList
                        If import of multiple SWCs, will return NeuronList of
                        TreeNeurons.

    See Also
    --------
    :func:`navis.to_swc`
                        Export neurons as SWC files.

    """
    if utils.is_iterable(f):
        return core.NeuronList([from_swc(x,
                                         connector_labels=connector_labels,
                                         include_subdirs=include_subdirs,
                                         **kwargs)
                                for x in config.tqdm(f, desc='Importing',
                                                     disable=config.pbar_hide,
                                                     leave=config.pbar_leave)])

    header = []
    if isinstance(f, pd.DataFrame):
        nodes = f
        f = 'SWC'
    elif isinstance(f, str) and os.path.isdir(f):
        if not include_subdirs:
            swc = [os.path.join(f, x) for x in os.listdir(f) if
                   os.path.isfile(os.path.join(f, x)) and x.endswith('.swc')]
        else:
            swc = [y for x in os.walk(f) for y in glob(os.path.join(x[0], '*.swc'))]

        return core.NeuronList([from_swc(x,
                                         connector_labels=connector_labels)
                                for x in config.tqdm(swc,
                                                     desc='Reading {}'.format(f.split('/')[-1]),
                                                     disable=config.pbar_hide,
                                                     leave=config.pbar_leave)])
    elif isinstance(f, str):
        data = []
        if os.path.isfile(f):
            with open(f) as file:
                reader = csv.reader(file, delimiter=' ')
                for row in reader:
                    # skip empty rows
                    if not row:
                        continue
                    # skip comments
                    if not row[0].startswith('#'):
                        data.append(row)
                    else:
                        header.append(' '.join(row))

        # If not file, assume it's a SWC string or a URL
        else:
            # Check if is url
            if utils.is_url(f):
                r = requests.get(f)
                r.raise_for_status()

                f = r.content.decode()

            # Note that with .split(), the last row will be empty
            rows = f.split('\n')[:-1]
            for r in rows:
                if not r.startswith('#'):
                    data.append(r.split(' '))
                else:
                    header.append(r)

            # Change f to generic name so that we can use it as name
            f = 'SWC'

        if not data:
            raise ValueError('No data found in SWC.')

        # Remove empty entries and generate nodes DataFrame
        nodes = pd.DataFrame([[e for e in row if e != ''] for row in data],
                             columns=['node_id', 'label', 'x', 'y', 'z',
                                      'radius', 'parent_id'],
                             dtype=object)
    else:
        raise TypeError('"f" must be filename, SWC string or DataFrame, not '
                        f'{type(f)}')

    # Turn header back into single string
    header = '\n'.join(header)

    # If any invalid nodes are found
    if np.any(nodes[['node_id', 'parent_id', 'x', 'y', 'z']].isnull()):
        # Remove nodes with missing data
        nodes = nodes.loc[~nodes[['node_id', 'parent_id', 'x', 'y', 'z']].isnull().any(axis=1)]

        # Because we removed nodes, we'll have to run a more complicated root
        # detection
        nodes.loc[~nodes.parent_id.isin(nodes.node_id), 'parent_id'] = -1

    # Convert data to respective dtypes
    dtypes = {'node_id': int, 'parent_id': int, 'label': str,
              'x': float, 'y': float, 'z': float, 'radius': float}

    for k, v in dtypes.items():
        if isinstance(v, type):
            nodes[k] = nodes[k].astype(v, errors='ignore')
        else:
            nodes[k] = nodes[k].map(v)
            nodes[k] = nodes[k].astype(object)

    # Take care of connectors
    if connector_labels:
        connectors = pd.DataFrame([], columns=['node_id', 'connector_id',
                                               'type', 'x', 'y', 'z'],
                                  dtype=object)
        for t, l in connector_labels.items():
            cn = nodes[nodes.label == l][['node_id', 'x', 'y', 'z']].copy()
            cn['connector_id'] = None
            cn['type'] = t
            connectors = pd.concat([connectors, cn], axis=0)
    else:
        connectors = None

    n = core.TreeNeuron(nodes,
                        connectors=connectors,
                        name=kwargs.pop('name',
                                        os.path.basename(f).replace('.swc', '')),
                        filename=os.path.basename(f),
                        pathname=os.path.dirname(f),
                        file=f,
                        header=header,
                        soma_label=soma_label,
                        connector_labels=connector_labels,
                        created_at=str(datetime.datetime.now()),
                        **kwargs)

    return n


def to_swc(x: 'core.NeuronObject',
           filename: Optional[str] = None,
           header: Optional[str] = None,
           labels: bool = True,
           export_synapses: bool = False) -> None:
    """Generate SWC file from neuron(s).

    Follows the format specified
    `here <http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html>`_.

    Parameters
    ----------
    x :                 TreeNeuron | NeuronList
                        If multiple neurons, will generate a single SWC file
                        for each neuron (see also ``filename``).
    filename :          None | str | list, optional
                        If ``None``, will use "neuron_{skeletonID}.swc". Pass
                        filenames as list when processing multiple neurons.
    header :            str | None, optional
                        Header for SWC file. If not provided, will use
                        generic header.
    labels :            str | dict | bool, optional
                        Node labels. Can be::

                            str : column name in node table
                            dict: must be of format {node_id: 'label', ...}.
                            bool: if True, will generate automatic labels, if False all nodes have label "0".

    export_connectors : bool, optional
                        If True, will label nodes with pre- ("7") and
                        postsynapse ("8"). Because only one label can be given
                        this might drop synapses (i.e. in case of multiple
                        pre- or postsynapses on a single node)! ``labels``
                        must be ``True`` for this to have any effect.

    Returns
    -------
    Nothing

    See Also
    --------
    :func:`navis.from_swc`
                        Import SWC files.

    """
    if isinstance(x, core.NeuronList):
        if not utils.is_iterable(filename):
            filename = [filename] * len(x)

        # At this point filename is iterable
        filename: Iterable[str]
        for n, f in zip(x, filename):
            to_swc(n, f,
                   export_synapses=export_synapses)
        return

    if not isinstance(x, core.TreeNeuron):
        raise ValueError('Can only process TreeNeurons, '
                         'got "{}"'.format(type(x)))

    # If not specified, generate generic filename
    if isinstance(filename, type(None)):
        filename = 'neuron_{}.swc'.format(x.skeleton_id)

    # Check if filename is of correct type
    if not isinstance(filename, str):
        raise ValueError('Filename must be str or None, '
                         'got "{}"'.format(type(filename)))

    # Make sure file ending is correct
    if os.path.isdir(filename):
        filename += 'neuron_{}.swc'.format(x.skeleton_id)
    elif not filename.endswith('.swc'):
        filename += '.swc'

    # Make copy of nodes and reorder such that the parent is always before a
    # node
    nodes_ordered = [n for seg in x.segments for n in seg[::-1]]
    this_tn = x.nodes.set_index('node_id', inplace=False).loc[nodes_ordered]

    # Because the last node ID of each segment is a duplicate
    # (except for the first segment ), we have to remove these
    this_tn = this_tn[~this_tn.index.duplicated(keep='first')]

    # Add an index column (must start with "1", not "0")
    this_tn['index'] = list(range(1, this_tn.shape[0] + 1))

    # Make a dictionary node_id -> index
    tn2ix = this_tn['index'].to_dict()

    # Make parent index column
    this_tn['parent_ix'] = this_tn.parent_id.map(lambda x: tn2ix.get(x, -1))

    # Add labels
    if isinstance(labels, dict):
        this_tn['label'] = this_tn.index.map(labels)
    elif isinstance(labels, str):
        this_tn['label'] = this_tn[labels]
    else:
        this_tn['label'] = 0
        # Add end/branch labels
        this_tn.loc[this_tn.type == 'branch', 'label'] = 5
        this_tn.loc[this_tn.type == 'end', 'label'] = 6
        # Add soma label
        if x.soma:
            this_tn.loc[x.soma, 'label'] = 1
        if export_synapses:
            # Add synapse label
            this_tn.loc[x.presynapses.node_id.values, 'label'] = 7
            this_tn.loc[x.postsynapses.node_id.values, 'label'] = 8

    # Generate table consisting of PointNo Label X Y Z Radius Parent
    # .copy() is to prevent pandas' chaining warnings
    swc = this_tn[['index', 'label', 'x', 'y', 'z',
                   'radius', 'parent_ix']].copy()

    # Adjust column titles
    swc.columns = ['PointNo', 'Label', 'X', 'Y', 'Z', 'Radius', 'Parent']

    with open(filename, 'w') as file:
        # Write header
        if not isinstance(header, str):
            file.write('# SWC format file\n')
            file.write('# based on specifications at http://research.mssm.edu/cnic/swc.html\n')
            file.write('# Created on {} using navis (https://github.com/schlegelp/navis)\n'.format(str(datetime.date.today())))
            file.write('# PointNo Label X Y Z Radius Parent\n')
            file.write('# Labels:\n')
            for l in ['0 = undefined', '1 = soma', '5 = fork point', '6 = end point']:
                file.write('# {}\n'.format(l))
            if export_synapses:
                for l in ['7 = presynapse', '8 = postsynapse']:
                    file.write('# {}\n'.format(l))
        else:
            file.write(header)

        writer = csv.writer(file, delimiter=' ')
        writer.writerows(swc.astype(str).values)


def to_float(x: Any) -> Optional[float]:
    """ Helper to try to convert to float.
    """
    try:
        return float(x)
    except BaseException:
        return None


def to_int(x: Any) -> Optional[int]:
    """ Helper to try to convert to float.
    """
    try:
        return int(x)
    except BaseException:
        return None



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import networkx as nx
import math
import pandas as pd
import scipy.spatial

from typing import Union, Optional, List, Iterable

try:
    import igraph
except ImportError:
    igraph = None

from .. import config, core

# Set up logging
logger = config.logger

__all__ = sorted(['network2nx', 'network2igraph', 'neuron2igraph', 'nx2neuron',
                  'neuron2nx', 'neuron2KDTree', 'neuron2dps'])


def network2nx(x: Union[pd.DataFrame, Iterable],
               threshold: int = 1,
               group_by: Union[dict, None] = None) -> nx.DiGraph:
    """Generates NetworkX graph for neuron connectivity.

    Parameters
    ----------
    x :                 pandas.DataFrame
                        Connectivity information:

                         1. List of edges (columns: 'source', 'target', 'weight')
                         2. Adjacency matrix (pd.DataFrame, rows=sources,
                            columns=targets)

    threshold :         int, optional
                        Connections weaker than this will be excluded.
    group_by :          None | dict, optional
                        Provide a dictionary ``{group_name: [skid1, skid2, ...]}``
                        to collapse sets of nodes into groups.

    Returns
    -------
    networkx.DiGraph
                        NetworkX representation of the network.

    """

    if isinstance(x, pd.DataFrame):
        miss = [c not in x.columns for c in ['source', 'target', 'weight']]
        if all(miss):
            edges = x[['source', 'target', 'weight']].values
        else:
            edges = x.reset_index(inplace=False,
                                  drop=False).melt(id_vars='index',
                                                   inplace=False).values
    elif isinstance(x, (list, np.ndarray)):
        if any([len(e) != 3 for e in x]):
            raise ValueError('Edges must be [source, target, weight]')
        edges = x

    edges = np.array(edges)

    if threshold:
        edges = edges[edges[:, 2] >= threshold]

    # Generate graph and assign custom properties
    g = nx.DiGraph()
    g.add_weighted_edges_from(edges)

    # Group nodes
    if group_by:
        for n, skids in group_by.items():
            # First collapse all nodes into the first of each group
            for s in skids[1:]:
                g = nx.contracted_nodes(g, str(skids[0]), str(s))
            # Now relabel the first node
            g = nx.relabel_nodes(g, {str(skids[0]): str(n)})
            g.nodes[str(n)]['neuron_name'] = str(n)

    return g


def network2igraph(x: Union[pd.DataFrame, Iterable],
                   threshold: int = 1) -> 'igraph.Graph':
    """Generate iGraph graph for neuron connectivity.

    Requires iGraph to be installed.

    Parameters
    ----------
    x :                 pandas.DataFrame
                        Connectivity information:

                         1. List of edges (columns: 'source', 'target', 'weight')
                         2. Adjacency matrix (pd.DataFrame, rows=sources,
                            columns=targets)

    threshold :         int, optional
                        Connections weaker than this will be excluded .

    Returns
    -------
    igraph.Graph(directed=True)
                        iGraph representation of the network.

    Examples
    --------
    >>> import navis
    >>> import igraph
    >>> g = navis.network2igraph('annotation:large network',
    ...                           remote_instance=rm)
    >>> # Plot graph
    >>> igraph.plot(g)
    >>> # Plot with edge width
    >>> igraph.plot(g, **{'edge_width': [ w/10 for w in g.es['weight'] ] })
    >>> # Plot with edge label
    >>> igraph.plot(g, **{'edge_label': g.es['weight'] })
    >>> # Save as graphml to import into e.g. Cytoscape
    >>> g.save('graph.graphml')

    """
    if igraph is None:
        raise ImportError('igraph must be installed to use this function.')

    if isinstance(x, pd.DataFrame):
        miss = [c not in x.columns for c in ['source', 'target', 'weight']]
        if all(miss):
            edges = x[['source', 'target', 'weight']].values
        else:
            edges = x.reset_index(inplace=False,
                                  drop=False).melt(id_vars='index',
                                                   inplace=False).values
    elif isinstance(x, (list, np.ndarray)):
        if any([len(e) != 3 for e in x]):
            raise ValueError('Edges must be [source, target, weight]')
        edges = x

    edges = np.array(edges)

    if threshold:
        edges = edges[edges[:, 2] >= threshold]

    names = list(set(np.array(edges)[:, 0]) & set(np.array(edges)[:, 1]))

    edges_by_index = [[names.index(e[0]), names.index(e[1])] for e in edges]

    # Generate igraph and assign custom properties
    g = igraph.Graph(directed=True)
    g.add_vertices(len(names))
    g.add_edges(edges_by_index)

    g.vs['node_id'] = names
    # g.vs['neuron_name'] = g.vs['label'] = neuron_names
    g.es['weight'] = edges[:, 2]

    return g


def neuron2nx(x: 'core.NeuronObject') -> nx.DiGraph:
    """ Turn TreeNeuron into an NetworkX DiGraph.

    Parameters
    ----------
    x :         TreeNeuron | NeuronList

    Returns
    -------
    networkx.DiGraph
                NetworkX representation of the neuron. Returns list of graphs
                if x is multiple neurons.

    """

    if isinstance(x, (pd.DataFrame, core.NeuronList)):
        return [neuron2nx(x.loc[i]) for i in range(x.shape[0])]
    elif isinstance(x, (pd.Series, core.TreeNeuron)):
        pass
    else:
        raise ValueError(f'Wrong input type "{type(x)}"')

    # Collect nodes
    nodes = x.nodes.set_index('node_id', inplace=False)
    # Collect edges
    edges = x.nodes[x.nodes.parent_id >= 0][['node_id', 'parent_id']].values
    # Collect weight
    weights = np.sqrt(np.sum((nodes.loc[edges[:, 0], ['x', 'y', 'z']].values.astype(float) -
                              nodes.loc[edges[:, 1], ['x', 'y', 'z']].values.astype(float)) ** 2, axis=1))
    # Generate weight dictionary
    edge_dict = np.array([{'weight': w} for w in weights])
    # Add weights to dictionary
    edges = np.append(edges, edge_dict.reshape(len(edges), 1), axis=1)
    # Create empty directed Graph
    g = nx.DiGraph()
    # Add nodes (in case we have disconnected nodes)
    g.add_nodes_from(x.nodes.node_id.values)
    # Add edges
    g.add_edges_from(edges)

    return g


def neuron2igraph(x: 'core.NeuronObject') -> 'igraph.Graph':
    """ Turns TreeNeuron(s) into an iGraph graph.

    Requires iGraph to be installed.

    Parameters
    ----------
    x :         TreeNeuron | NeuronList

    Returns
    -------
    igraph.Graph(directed=True)
                Representation of the neuron. Returns list of graphs
                if x is multiple neurons.
    None
                If igraph not installed.

    """
    # If iGraph is not installed return nothing
    if igraph is None:
        return None

    if isinstance(x, (pd.DataFrame, core.NeuronList)):
        return [neuron2igraph(x.loc[i]) for i in range(x.shape[0])]
    elif isinstance(x, (pd.Series, core.TreeNeuron)):
        pass
    else:
        raise ValueError(f'Unable input type "{type(x)}"')

    logger.debug('Generating graph from skeleton data...')

    # Make sure we have correctly numbered indices
    nodes = x.nodes.reset_index(inplace=False, drop=True)

    # Generate list of vertices -> this order is retained
    vlist = nodes.node_id.values

    # Get list of edges as indices (needs to exclude root node)
    tn_index_with_parent = nodes.index.values[nodes.parent_id >= 0]
    parent_ids = nodes.parent_id.values[nodes.parent_id >= 0]
    nodes['temp_index'] = nodes.index  # add temporary index column
    parent_index = nodes.set_index('node_id', inplace=False).loc[parent_ids,
                                                                 'temp_index'].values

    # Generate list of edges based on index of vertices
    elist = np.vstack((tn_index_with_parent, parent_index)).T

    # iGraph < 0.8.0 does not like arrays as edge list
    if getattr(igraph, '__version_info__', (0, 0, 0))[1] < 8:
        elist = elist.tolist()

    # Generate graph and assign custom properties
    g = igraph.Graph(elist, n=len(vlist), directed=True)

    g.vs['node_id'] = g.vs['name'] = nodes.node_id.values
    g.vs['parent_id'] = nodes.parent_id.values

    # Generate weights by calculating edge lengths = distance between nodes
    tn_coords = nodes[['x', 'y', 'z']].values[tn_index_with_parent, :]
    parent_coords = nodes[['x', 'y', 'z']].values[parent_index.astype(int), :]

    w = np.sqrt(np.sum((tn_coords - parent_coords) ** 2, axis=1))
    g.es['weight'] = w

    return g


def nx2neuron(g: nx.Graph,
              root: Optional[Union[int, str]] = None
              ) -> pd.DataFrame:
    """Generate node table from NetworkX Graph.

    This function will try to generate a neuron-like tree structure from
    the Graph. Therefore the graph may not contain loops!

    Node attributes (e.g. ``x``, ``y``, ``z``, ``radius``, ``confidence``) need
    to be properties of the graph's nodes. All node property will be added to
    the neuron's ``.nodes`` table.

    Parameters
    ----------
    g :             networkx.Graph
    root :          str | int, optional
                    Node in graph to use as root for neuron. If not provided,
                    will use first node in ``g.nodes``.

    Returns
    -------
    pandas.DataFrame

    """
    # First some sanity checks
    if not isinstance(g, nx.Graph):
        raise TypeError(f'g must be NetworkX Graph, not "{type(g)}"')

    # We need an undirected Graph
    if isinstance(g, nx.DiGraph):
        g = g.to_undirected(as_view=True)

    if not nx.is_tree(g):
        raise TypeError("g must be tree-like. Please check for loops.")

    # Pick a randon root if not explicitly provided
    if not root:
        root = list(g.nodes)[0]
    elif root not in g.nodes:
        raise ValueError(f'Node "{root}" not in graph.')

    # Generate parent->child dictionary
    lop = nx.predecessor(g, root)

    # Make sure no node has more than one parent
    if max([len(v) for v in lop.values()]) > 1:
        raise ValueError('Nodes with multiple parents found. Make sure graph '
                         'is tree-like.')

    # Note that we assign -1 as root's parent
    lop = {k: v[0] if v else -1 for k, v in lop.items()}

    # Generate node table
    tn_table = pd.DataFrame(index=list(g.nodes))
    tn_table.index = tn_table.index.set_names('node_id', inplace=False)

    # Add parents - use -1 for root's parent
    tn_table['parent_id'] = tn_table.index.map(lop)

    try:
        tn_table.index = tn_table.index.astype(int)
        tn_table['parent_id'] = tn_table.parent_id.astype(int)
    except (ValueError, TypeError):
        raise ValueError('Node IDs must be convertible to integers.')
    except BaseException:
        raise

    # Add additional generic attribute -> will skip node_id and parent_id
    # if they exist
    all_attr = set([k for n in g.nodes for k in g.nodes[n].keys()])

    # Remove some that we don't need
    all_attr -= set(['parent_id', 'node_id'])
    # Add some that we want as columns even if they don't exist
    all_attr |= set(['x', 'y', 'z', 'confidence', 'radius'])

    # For some we want to have set default values
    defaults = {'x': 0, 'y': 0, 'z': 0, 'confidence': 5, 'radius': -1}

    # Now map the attributes onto node table
    for at in all_attr:
        vals = nx.get_node_attributes(g, at)
        tn_table[at] = tn_table.index.map(lambda a: vals.get(a, defaults.get(at)))

    return tn_table.reset_index(drop=False, inplace=False)


def _find_all_paths(g: nx.DiGraph,
                    start,
                    end,
                    mode: str = 'OUT',
                    maxlen: Optional[int] = None) -> list:
    """ Find all paths between two vertices in an iGraph object. For some reason
    this function exists in R iGraph but not Python iGraph. This is rather slow
    and should not be used for large graphs.
    """

    def find_all_paths_aux(adjlist: List[set],
                           start: int,
                           end: int,
                           path: list,
                           maxlen: Optional[int] = None) -> list:
        path = path + [start]
        if start == end:
            return [path]
        paths: list = []
        if maxlen is None or len(path) <= maxlen:
            for node in adjlist[start] - set(path):
                paths.extend(find_all_paths_aux(adjlist,
                                                node,
                                                end,
                                                path,
                                                maxlen))
        return paths

    adjlist = [set(g.neighbors(node, mode=mode))
               for node in range(g.vcount())]
    all_paths: list = []
    start = start if isinstance(start, list) else [start]
    end = end if isinstance(end, list) else [end]
    for s in start:
        for e in end:
            all_paths.extend(find_all_paths_aux(adjlist, s, e, [], maxlen))
    return all_paths


def neuron2KDTree(x: 'core.TreeNeuron',
                  tree_type: str = 'c',
                  data: str = 'nodes',
                  **kwargs) -> Union[scipy.spatial.cKDTree,
                                     scipy.spatial.KDTree]:
    """ Turns a neuron into scipy KDTree.

    Parameters
    ----------
    x :         single TreeNeuron
    tree_type : 'c' | 'normal', optional
                Type of KDTree:
                  1. ``'c'`` = ``scipy.spatial.cKDTree`` (faster)
                  2. ``'normal'`` = ``scipy.spatial.KDTree`` (more functions)
    data :      'nodes' | 'connectors', optional
                Data to use to generate tree.
    **kwargs
                Keyword arguments passed at KDTree initialization.


    Returns
    -------
    ``scipy.spatial.cKDTree`` or ``scipy.spatial.KDTree``

    """

    if tree_type not in ['c', 'normal']:
        raise ValueError('"tree_type" needs to be either "c" or "normal"')

    if data not in ['nodes', 'connectors']:
        raise ValueError('"data" needs to be either "nodes" or "connectors"')

    if isinstance(x, core.NeuronList):
        if len(x) == 1:
            x = x[0]
        else:
            raise ValueError('Need a single TreeNeuron')
    elif not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Need TreeNeuron, got "{type(x)}"')

    if data == 'nodes':
        d = x.nodes[['x', 'y', 'z']].values
    else:
        d = x.connectors[['x', 'y', 'z']].values

    if tree_type == 'c':
        return scipy.spatial.cKDTree(data=d, **kwargs)
    else:
        return scipy.spatial.KDTree(data=d, **kwargs)


def neuron2dps(x: 'core.TreeNeuron') -> pd.DataFrame:
    """Convert neuron to point clouds with tangent vectors (but no connectivity).

    Dotprops consist of a point and a vector. This works by (1) finding the
    center between child->parent nodes and (2) getting the vector between
    them. Also returns the length of the vector.

    Parameters
    ----------
    x :         TreeNeuron
                Single neuron

    Returns
    -------
    pandas.DataFrame
            DataFrame in which each row represents a segments between two
            nodes::

                point  vector  vec_length
             1
             2
             3

    Examples
    --------
    >>> import navis
    >>> x = navis.example_neurons()
    >>> dps = navis.neuron2dps(x)
    >>> # Get array of all locations
    >>> locs = numpy.vstack(dps.point.values)

    See Also
    --------
    navis.TreeNeuron.dps
            Shorthand to the dotprops representation of neuron.

    """

    if isinstance(x, core.NeuronList):
        if x.shape[0] == 1:
            x = x[0]
        else:
            raise ValueError('Please pass only single TreeNeurons')

    if not isinstance(x, core.TreeNeuron):
        raise ValueError('Can only process TreeNeurons')

    # First, get a list of child -> parent locs (exclude root node!)
    tn_locs = x.nodes[x.nodes.parent_id >= 0][['x', 'y', 'z']].values
    pn_locs = x.nodes.set_index('node_id', inplace=False).loc[x.nodes[x.nodes.parent_id >= 0].parent_id][['x', 'y', 'z']].values

    # Get centers between each pair of locs
    centers = tn_locs + (pn_locs - tn_locs) / 2

    # Get vector between points
    vec = pn_locs - tn_locs

    dps = pd.DataFrame([[c, v] for c, v in zip(centers, vec)],
                       columns=['point', 'vector'])

    # Add length of vector (for convenience)
    dps['vec_length'] = (dps.vector ** 2).apply(sum).apply(math.sqrt)

    return dps



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from .mmetrics import (strahler_index, bending_flow, flow_centrality,
                      segregation_index, tortuosity)
from .manipulation import (prune_by_strahler, stitch_neurons,
                           split_axon_dendrite, average_neurons,
                           despike_neuron, guess_radius, smooth_neuron,
                           heal_fragmented_neuron, break_fragments,
                           prune_twigs)
from .analyze import find_soma



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


""" This module contains functions for intersections.
"""

import pandas as pd
import numpy as np

from typing import Union, List, Dict, Sequence, Optional, overload, Any
from typing_extensions import Literal

from .. import config, graph, core

from .ray import *
from .convex import *

# Set up logging -> has to be before try statement!
logger = config.logger

try:
    from pyoctree import pyoctree
except ImportError:
    pyoctree = None
    logger.warning("Module pyoctree not found. Falling back to scipy's"
                   "ConvexHull for intersection calculations.")

__all__ = sorted(['in_volume', 'intersection_matrix'])


@overload
def in_volume(x: 'core.NeuronObject',
              volume: core.Volume,
              inplace: Literal[True],
              mode: Union[Literal['IN'], Literal['OUT']] = 'IN',
              method: Union[Literal['FAST'], Literal['SAFE']] = 'FAST',
              prevent_fragments: bool = False) -> None: ...


@overload
def in_volume(x: 'core.TreeNeuron',
              volume: core.Volume,
              inplace: Literal[False],
              mode: Union[Literal['IN'], Literal['OUT']] = 'IN',
              method: Union[Literal['FAST'], Literal['SAFE']] = 'FAST',
              prevent_fragments: bool = False) -> 'core.TreeNeuron': ...


@overload
def in_volume(x: 'core.NeuronList',
              volume: core.Volume,
              inplace: Literal[False],
              mode: Union[Literal['IN'], Literal['OUT']] = 'IN',
              method: Union[Literal['FAST'], Literal['SAFE']] = 'FAST',
              prevent_fragments: bool = False) -> 'core.NeuronList': ...


@overload
def in_volume(x: Union[Sequence, pd.DataFrame],
              volume: core.Volume,
              inplace: bool = False,
              mode: Union[Literal['IN'], Literal['OUT']] = 'IN',
              method: Union[Literal['FAST'], Literal['SAFE']] = 'FAST',
              prevent_fragments: bool = False) -> Sequence[bool]: ...


@overload
def in_volume(x: Union['core.NeuronObject', Sequence, pd.DataFrame],
              volume: Union[Dict[str, core.Volume],
                            Sequence[core.Volume]],
              inplace: bool = False,
              mode: Union[Literal['IN'], Literal['OUT']] = 'IN',
              method: Union[Literal['FAST'], Literal['SAFE']] = 'FAST',
              prevent_fragments: bool = False) -> Dict[str,
                                                       Union[Sequence[bool],
                                                             'core.NeuronObject']]: ...


# We do need the full signature b/c we're recursively calling in_volume
# which means that there is uncertainty
@overload
def in_volume(x: Union['core.NeuronObject', Sequence, pd.DataFrame],
              volume: Union[core.Volume,
                            Dict[str, core.Volume],
                            Sequence[core.Volume]],
              inplace: bool = False,
              mode: Union[Literal['IN'], Literal['OUT']] = 'IN',
              method: Union[Literal['FAST'], Literal['SAFE']] = 'FAST',
              prevent_fragments: bool = False) -> Optional[Union['core.NeuronObject',
                                                                 Sequence[bool],
                                                                 Dict[str, Union[Sequence[bool],
                                                                                 'core.NeuronObject']]
                                                                 ]]: ...


def in_volume(x: Union['core.NeuronObject', Sequence, pd.DataFrame],
              volume: Union[core.Volume,
                            Dict[str, core.Volume],
                            Sequence[core.Volume]],
              inplace: bool = False,
              mode: Union[Literal['IN'], Literal['OUT']] = 'IN',
              method: Union[Literal['FAST'], Literal['SAFE']] = 'FAST',
              prevent_fragments: bool = False) -> Optional[Union['core.NeuronObject',
                                                                 Sequence[bool],
                                                                 Dict[str, Union[Sequence[bool],
                                                                                 'core.NeuronObject']]
                                                                 ]]:
    """ Test if points/neurons are within a given volume.

    Important
    ---------
    This function requires `pyoctree <https://github.com/mhogg/pyoctree>`_
    which is only an optional dependency of navis. If pyoctree is not
    installed, we will fall back to using scipy ConvexHull instead. This is
    slower and may give wrong positives for concave meshes!

    Parameters
    ----------
    x :                 list of tuples | numpy.array | pandas.DataFrame | TreeNeuron | NeuronList

                        - list/numpy.array is treated as list of x/y/z
                          coordinates. Needs to be shape (N,3): e.g.
                          ``[[x1, y1, z1], [x2, y2, z2], ..]``
                        - ``pandas.DataFrame`` needs to have ``x, y, z``
                          columns

    volume :            navis.Volume | dict or list of navis.Volume
                        :class:`navis.Volume` to test. Multiple volumes can
                        be given as list (``[volume1, volume2, ...]``) or dict
                        (``{'label1': volume1, ...}``).
    inplace :           bool, optional
                        If False, a copy of the original DataFrames/Neuron is
                        returned. Does only apply to TreeNeuron or
                        NeuronList objects. Does apply if multiple
                        volumes are provided.
    mode :              'IN' | 'OUT', optional
                        If 'IN', parts of the neuron that are within the volume
                        are kept.
    method :            'FAST' | 'SAFE', optional
                        Method used for raycasting. "FAST" will cast only a
                        single ray to check for intersections. If you
                        experience problems, set method to "SAFE" to use
                        multiple rays (slower).
    prevent_fragments : bool, optional
                        Only relevant if input is Neuron/List: if True,
                        will add nodes required to keep neuron from
                        fragmenting.

    Returns
    -------
    TreeNeuron
                      If input is TreeNeuron or NeuronList, will
                      return subset of the neuron(s) (nodes and connectors)
                      that are within given volume.
    list of bools
                      If input is a set of coordinates, returns boolean:
                      ``True`` if in volume, ``False`` if not in order.
    dict
                      If multiple volumes are provided, results will be
                      returned in dictionary with volumes as keys::

                        {'volume1': in_volume(x, volume1),
                         'volume2': in_volume(x, volume2),
                         ... }

    Examples
    --------
    # Prune neuron to volume

    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> lh = navis.example_volume('LH')
    >>> n_lh = navis.in_volume(n, lh, inplace=False)
    >>> n_lh
    type              TreeNeuron
    name            neuron_38885
    n_nodes                 1299
    n_connectors               0
    n_branches                54
    n_leafs                   56
    cable_length          294836
    soma                    None
    dtype: object


    """

    # If we are given multiple volumes
    if isinstance(volume, (list, dict, np.ndarray)):
        # Force into dict
        if not isinstance(volume, dict):
            # Make sure all Volumes can be uniquely indexed
            vnames = [v.name for v in volume if isinstance(v, core.Volume)]
            dupli = [v for v in set(vnames) if vnames.count(v) > 1]
            if dupli:
                raise ValueError('Duplicate Volume names detected: '
                                 f'{",".join(dupli)}. Volume.name must be '
                                 'unique.')

            volume = {v.name: v for v in volume if isinstance(v, core.Volume)}

        data: Dict[str, Any] = dict()
        for v in config.tqdm(volume, desc='Volumes', disable=config.pbar_hide,
                             leave=config.pbar_leave):
            data[v] = in_volume(x,
                                volume=volume[v],
                                inplace=False,
                                mode=mode,
                                method=method)
        return data

    # From here on out volume is a single core.Volume
    vol: 'core.Volume' = volume  # type: ignore

    # Make copy if necessary
    if isinstance(x, (core.NeuronList, core.TreeNeuron)):
        if inplace is False:
            x = x.copy()

    if isinstance(x, core.TreeNeuron):
        in_v = in_volume(x.nodes[['x', 'y', 'z']],
                         vol,
                         method=method)

        # If mode is OUT, invert selection
        if mode == 'OUT':
            in_v = ~np.array(in_v)

        x2 = graph.subset_neuron(x,
                                 subset=x.nodes[in_v].node_id.values,
                                 inplace=inplace,
                                 prevent_fragments=prevent_fragments)

        if inplace is False:
            return x2
        return None
    elif isinstance(x, core.NeuronList):
        for n in config.tqdm(x, desc='Subsetting',
                             leave=config.pbar_leave,
                             disable=config.pbar_hide):
            in_volume(n, vol, inplace=True, mode=mode, method=method,
                      prevent_fragments=prevent_fragments)

        if inplace is False:
            return x
        return None
    elif isinstance(x, pd.DataFrame):
        points = x[['x', 'y', 'z']].values
    elif isinstance(x, np.ndarray):
        points = x
    elif isinstance(x, (list, tuple)):
        points = np.array(x)

    if points.ndim != 2 or points.shape[1] != 3:  # type: ignore  # does not know about numpy
        raise ValueError('Points must be array of shape (N,3).')

    if pyoctree:
        return ray_in_volume(points, vol,
                             multi_ray=method.upper() == 'SAFE')
    else:
        logger.warning(
            'Package pyoctree not found. Falling back to ConvexHull.')
        return in_volume_convex(points, vol, approximate=False)


def intersection_matrix(x: 'core.NeuronObject',
                        volumes: Union[List[core.Volume],
                                       Dict[str, core.Volume]],
                        attr: Optional[str] = None,
                        method: Union[Literal['FAST'],
                                      Literal['SAFE']] = 'FAST'
                        ) -> pd.DataFrame:
    """ Computes intersection matrix between a set of neurons and a set of
    volumes.

    Parameters
    ----------
    x :               NeuronList | TreeNeuron
                      Neurons to intersect.
    volume :          list or dict of navis.Volume
    attr :            str | None, optional
                      Attribute to return for intersected neurons (e.g.
                      'cable_length'). If None, will return TreeNeuron.
    method :          'FAST' | 'SAFE', optional
                      See :func:`navis.intersect.in_volume`.

    Returns
    -------
    pandas DataFrame
    """

    # Volumes should be a dict at some point
    volumes_dict: Dict[str, core.Volume]

    if isinstance(x, core.TreeNeuron):
        x = core.NeuronList(x)

    if not isinstance(x, core.NeuronList):
        raise TypeError(f'x must be Neuron/List, not "{type(x)}"')

    if not isinstance(volumes, (list, dict)):
        raise TypeError('Volumes must be given as list or dict, not '
                        f'"{type(volumes)}"')

    if isinstance(volumes, list):
        volumes_dict = {v.name: v for v in volumes}
    else:
        volumes_dict = volumes

    for v in volumes_dict.values():
        if not isinstance(v, core.Volume):
            raise TypeError(f'Wrong data type found in volumes: "{type(v)}"')

    data = in_volume(x, volumes_dict, inplace=False, mode='IN', method=method)

    if not attr:
        df = pd.DataFrame([[n for n in data[v]] for v in data],
                          index=list(data.keys()),
                          columns=x.skeleton_id)
    else:
        df = pd.DataFrame([[getattr(n, attr) for n in data[v]] for v in data],
                          index=list(data.keys()),
                          columns=x.skeleton_id)

    return df



from .iterables import make_iterable, make_non_iterable, is_iterable
from .misc import (is_jupyter, set_loggers, set_pbars, unpack_neurons,
                   set_default_connector_colors, parse_objects,
                   is_url, make_url)
from .validate import validate_options, validate_table
from .eval import eval_node_ids, eval_neurons, eval_id, eval_conditions



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


""" This module contains functions to analyse connectivity.
"""

import pandas as pd
import numpy as np
import scipy.spatial

from typing import Union, Dict
from typing_extensions import Literal

from ..core.neurons import TreeNeuron
from ..core.neuronlist import NeuronList
from .. import config

# Set up logging
logger = config.logger

NeuronObject = Union[TreeNeuron, NeuronList]


def cable_overlap(a: NeuronObject,
                  b: NeuronObject,
                  dist: float = 2,
                  method: Union[Literal['min'], Literal['max'], Literal['avg']] = 'min'
                  ) -> pd.DataFrame:
    """ Calculates the amount of cable of neuron A within distance of neuron B.

    Uses dotproduct representation of a neuron! It is recommended to
    resample neurons first.

    Parameters
    ----------
    a,b :       TreeNeuron | NeuronList
                Neuron(s) for which to compute cable within distance.
    dist :      int | float, optional
                Maximum distance.
    method :    'min' | 'max' | 'avg'
                Method by which to calculate the overlapping cable between
                two cables::

                  Assuming that neurons A and B have 300 and 150 um of cable
                  within given distances:

                    1. 'min' returns 150
                    2. 'max' returns 300
                    3. 'avg' returns 225

    Returns
    -------
    pandas.DataFrame
            Matrix in which neurons A are rows, neurons B are columns. Cable
            within distance is given in microns::

                        skidB1   skidB2  skidB3  ...
                skidA1    5        1        0
                skidA2    10       20       5
                skidA3    4        3        15
                ...

    See Also
    --------
    :func:`navis.resample_neuron`
                Use to resample neurons before calculating overlap.

    Examples
    --------
    >>> import navis
    >>> nl = navis.example_neurons(2)
    >>> # Resample to 1 micron
    >>> nl.resample(1000, inplace=True)
    >>> # Get overlapping cable within 2 microns
    >>> ol = navis.cable_overlap(nl, nl, dist=2000)

    """

    if not isinstance(a, (TreeNeuron, NeuronList)) \
       or not isinstance(b, (TreeNeuron, NeuronList)):
        raise TypeError('Need to pass CatmaidNeurons')

    if not isinstance(a, NeuronList):
        a = NeuronList(a)

    if not isinstance(b, NeuronList):
        b = NeuronList(b)

    allowed_methods = ['min', 'max', 'avg']
    if method not in allowed_methods:
        raise ValueError('Unknown method "{0}". Allowed methods: "{0}"'.format(
            method, ','.join(allowed_methods)))

    matrix = pd.DataFrame(np.zeros((a.shape[0], b.shape[0])),
                          index=a.name, columns=b.name)

    with config.tqdm(total=len(a), desc='Calc. overlap',
                     disable=config.pbar_hide,
                     leave=config.pbar_leave) as pbar:
        # Keep track of KDtrees
        trees: Dict[str, scipy.spatial.cKDTree] = {}
        for nA in a:
            # Get cKDTree for nA
            tA = trees.get(nA.name, None)
            if not tA:
                trees[nA.name] = tA = scipy.spatial.cKDTree(
                    np.vstack(nA.dps.point), leafsize=10)

            for nB in b:
                # Get cKDTree for nB
                tB = trees.get(nB.name, None)
                if not tB:
                    trees[nB.name] = tB = scipy.spatial.cKDTree(
                        np.vstack(nB.dps.point), leafsize=10)

                # Query nB -> nA
                distA, ixA = tA.query(np.vstack(nB.dps.point),
                                      k=1,
                                      distance_upper_bound=dist,
                                      n_jobs=-1
                                      )
                # Query nA -> nB
                distB, ixB = tB.query(np.vstack(nA.dps.point),
                                      k=1,
                                      distance_upper_bound=dist,
                                      n_jobs=-1
                                      )

                nA_in_dist = nA.dps.loc[ixA[distA != float('inf')]]
                nB_in_dist = nB.dps.loc[ixB[distB != float('inf')]]

                if nA_in_dist.empty:
                    overlap = 0
                elif method == 'avg':
                    overlap = (nA_in_dist.vec_length.sum() +
                               nB_in_dist.vec_length.sum()) / 2
                elif method == 'max':
                    overlap = max(nA_in_dist.vec_length.sum(),
                                  nB_in_dist.vec_length.sum())
                elif method == 'min':
                    overlap = min(nA_in_dist.vec_length.sum(),
                                  nB_in_dist.vec_length.sum())

                matrix.at[nA.name, nB.name] = overlap

            pbar.update(1)

    return matrix



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import json
import os

import networkx as nx

from typing import Union, Optional
from typing_extensions import Literal

from ..core.volumes import Volume
from ..core.neuronlist import NeuronList
from ..core.neurons import TreeNeuron
from ..io import *
from ..graph import *

__all__ = ['example_neurons', 'example_volume']

fp = os.path.dirname(__file__)

gml_path = os.path.join(fp, 'gml')
swc_path = os.path.join(fp, 'swc')
vols_path = os.path.join(fp, 'volumes')

gml = [f for f in os.listdir(gml_path) if f.endswith('.gml')]
swc = [f for f in os.listdir(swc_path) if f.endswith('.swc')]
vols = [f for f in os.listdir(vols_path) if f.endswith('.json')]

NeuronObject = Union[TreeNeuron, NeuronList]


def example_neurons(n: Optional[int] = None,
                    source: Union[Literal['swc'],
                                  Literal['gml']] = 'swc'
                    ) -> NeuronObject:
    """ Load example neuron(s).

    These example neurons are olfactory projection neurons from the DA2
    glomerulus that have been manually reconstructed from an EM volume of an
    entire fruit fly brain (see Zheng et al., Cell 2018).

    Parameters
    ----------
    n :         int, None, optional
                Number of neurons to return. If None, will return all example
                neurons.
    source :    'swc' | 'gml', optional
                Neurons can be generated from SWC files or GML graphs. This
                is only relevant for testing.

    Returns
    -------
    TreeNeuron
                If ``n=1``.
    NeuronList
                If ``n>1``.


    Examples
    --------
    Load a single neuron
    >>> import navis
    >>> n = navis.example_neurons(n=1)

    Load all example neurons
    >>> nl = navis.example_neurons()

    """

    if not isinstance(n, (int, type(None))):
        raise TypeError(f'Expected int or None, got "{type(n)}"')

    if isinstance(n, int) and n < 1:
        raise ValueError("Unable to return less than 1 neuron.")

    if source == 'gml':
        graphs = [nx.read_gml(os.path.join(gml_path, g)) for g in gml[:n]]
        nl = NeuronList([nx2neuron(g) for g in graphs], units='nm')
    elif source == 'swc':
        nl = NeuronList([from_swc(os.path.join(swc_path, f), units='nm') for f in swc[:n]])
    else:
        raise ValueError(f'Source must be "swc" or "gml", not "{source}"')

    # Invert the y-axis
    # This is necessary because CATMAID neurons have their origin
    # in the top left corner but e.g. matplotlib plots have it
    # in the bottom left corner
    for n in nl:
        n.nodes['y'] *= -1

    if n == 1:
        return nl[0]
    return nl


def example_volume(name: str) -> Volume:
    """ Load an example volume.

    Parameters
    ----------
    name :      str
                Name of available volume. Currently available::
                  "LH" = lateral horn in FAFB v14 space
                  'neuropil' = neuropil of FAFB v14 volume


    Returns
    -------
    navis.Volume

    Examples
    --------
    Load LH volume
    >>> import navis
    >>> lh = navis.example_volume('LH')
    """
    if not isinstance(name, str):
        raise TypeError(f'Expected string, got "{type(name)}"')

    # Force lower case
    name = name.lower()

    # Make sure extension is correct
    if not name.endswith('.json'):
        name += '.json'

    if name not in vols:
        raise ValueError(f'No volume named "{name}". Available volumes: {",".join(vols)}')

    with open(os.path.join(vols_path, name), 'r') as f:
        data = json.load(f)
        return Volume(**data)



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Module contains functions to plot neurons in 2D/2.5D.
"""
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import mpl_toolkits
from mpl_toolkits.mplot3d.art3d import Line3DCollection
from mpl_toolkits.mplot3d import proj3d
from matplotlib.collections import LineCollection

import numpy as np
import pint
import warnings

from typing import Union, List, Tuple
from typing_extensions import Literal

from .. import utils, config, core
from .colors import prepare_colormap
from .plot_utils import segments_to_coords, tn_pairs_to_coords

__all__ = ['plot2d']

logger = config.logger

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    pint.Quantity([])


def plot2d(x: Union[core.NeuronObject,
                    core.Volume,
                    np.ndarray,
                    List[Union[core.NeuronObject, np.ndarray, core.Volume]]
                    ],
           method: Union[Literal['2d'],
                         Literal['3d'],
                         Literal['3d_complex']] = '2d',
           **kwargs) -> Tuple[mpl.figure.Figure, mpl.axes.Axes]:
    """ Generate 2D plots of neurons and neuropils.

    The main advantage of this is that you can save plot as vector graphics.

    Important
    ---------
    This function uses matplotlib which "fakes" 3D as it has only very limited
    control over layers. Therefore neurites aren't necessarily plotted in the
    right Z order which becomes especially troublesome when plotting a complex
    scene with lots of neurons criss-crossing. See the ``method`` parameter
    for details. All methods use orthogonal projection.

    Parameters
    ----------
    x :               skeleton IDs | TreeNeuron | NeuronList | Volume | Dotprops | np.ndarray
                      Objects to plot::

                        - int is intepreted as skeleton ID(s)
                        - str is intepreted as volume name(s)
                        - multiple objects can be passed as list (see examples)
                        - numpy array of shape (n,3) is intepreted as scatter
    method :          '2d' | '3d' | '3d_complex'
                      Method used to generate plot. Comes in three flavours:
                        1. '2d' uses normal matplotlib. Neurons are plotted in
                           the order their are provided. Well behaved when
                           plotting neuropils and connectors. Always gives
                           frontal view.
                        2. '3d' uses matplotlib's 3D axis. Here, matplotlib
                           decide the order of plotting. Can chance perspective
                           either interacively or by code (see examples).
                        3. '3d_complex' same as 3d but each neuron segment is
                           added individually. This allows for more complex
                           crossing patterns to be rendered correctly. Slows
                           down rendering though.

    **kwargs
                      See Notes for permissible keyword arguments.

    Examples
    --------
    >>> import navis
    >>> import matplotlib.pyplot as plt

    Plot list of neurons as simple 2d
    >>> nl = navis.example_neurons()
    >>> fig, ax = navis.plot2d(nl)
    >>> plt.show()

    Add a volume
    >>> v = navis.example_volume('LH')
    >>> fig, ax = navis.plot2d([nl, vol])
    >>> plt.show()

    Change neuron colors
    >>> fig, ax = navis.plot2d(nl, color=['r', 'g', 'b', 'm', 'c', 'y'])
    >>> plt.show()

    Plot in "fake" 3D
    >>> fig, ax = navis.plot2d(nl, method='3d')
    >>> plt.show()
    >>> # Try dragging the window

    Plot in "fake" 3D and change perspective
    >>> fig, ax = navis.plot2d(nl, method='3d')
    >>> # Change view to lateral
    >>> ax.azim = 0
    >>> ax.elev = 0
    >>> # Change view to top
    >>> ax.azim = -90
    >>> ax.elev = 90
    >>> # Tilted top view
    >>> ax.azim = -135
    >>> ax.elev = 45
    >>> # Move camera closer (will make image bigger)
    >>> ax.dist = 5
    >>> plt.show()

    Plot using depth-coloring
    >>> fig, ax = navis.plot2d(nl, method='3d', depth_coloring=True)
    >>> plt.show()


    Returns
    --------
    fig, ax :      matplotlib figure and axis object

    Notes
    -----

    Optional keyword arguments:

    ``soma`` (bool, default = True)
       Plot soma if one exists.

    ``connectors`` (boolean, default = True)
       Plot connectors (synapses, gap junctions, abutting)

    ``connectors_only`` (boolean, default = False)
       Plot only connectors, not the neuron.

    ``cn_size`` (int | float, default = 1)
      Size of connectors.

    ``linewidth``/``lw`` (int | float, default = .5)
      Width of neurites.

    ``linestyle``/``ls`` (str, default = '-')
      Line style of neurites.

    ``autoscale`` (bool, default=True)
       If True, will scale the axes to fit the data.

    ``scalebar`` (int | float | str | pint.Quantity, default=False)
       Adds scale bar. Provide integer, float or str to set size of scalebar.
       Int|float are assumed to be in same units as data. You can specify
       units in as string: e.g. "1 um".
       For methods '3d' and '3d_complex', this will create an axis object.

    ``ax`` (matplotlib ax, default=None)
       Pass an ax object if you want to plot on an existing canvas.

    ``figsize`` (tuple, default = (8, 8))
      Size of figure.

    ``color`` (tuple | list | str | dict)
      Tuples/lists (r,g,b) and str (color name) are interpreted as a single
      colors that will be applied to all neurons. Dicts will be mapped onto
      neurons by skeleton ID.

    ``alpha`` (float [0-1], default = .9)
      Alpha value for neurons. Overriden if alpha is provided as fourth value
      in ``color``.

    ``use_neuron_color`` (bool, default = False)
      If True, will attempt to use ``.color`` attribute of neurons.

    ``depth_coloring`` (bool, default = False)
      If True, will color encode depth (Z). Overrides ``color``. Does not work
      with ``method = '3d_complex'``.

    ``depth_scale`` (bool, default = True)
      If True and ``depth_coloring=True`` will plot a scale.

    ``cn_mesh_colors`` (bool, default = False)
      If True, will use the neuron's color for its connectors too.

    ``group_neurons`` (bool, default = False)
      If True, neurons will be grouped. Works with SVG export (not PDF).
      Does NOT work with ``method='3d_complex'``.

    ``scatter_kws`` (dict, default = {})
      Parameters to be used when plotting points. Accepted keywords are:
      ``size`` and ``color``.

    ``view`` (tuple, default = ("x", "y"))
      Sets view for ``method='2d'``.

    ``volume_outlines`` (bool, default=True)
      If True will plot volume outline with no fill.


    See Also
    --------
    :func:`navis.plot3d`
            Use this if you want interactive, perspectively correct renders
            and if you don't need vector graphics as outputs.
    :func:`navis.plot1d`
            A nifty way to visualise neurons in a single dimension.

    """

    # Filter kwargs
    _ACCEPTED_KWARGS = ['soma', 'connectors', 'connectors_only',
                        'ax', 'color', 'colors', 'c', 'view', 'scalebar',
                        'cn_mesh_colors', 'linewidth', 'cn_size',
                        'group_neurons', 'scatter_kws', 'figsize', 'linestyle',
                        'alpha', 'depth_coloring', 'autoscale', 'depth_scale',
                        'use_neuron_color', 'ls', 'lw', 'volume_outlines']
    wrong_kwargs = [a for a in kwargs if a not in _ACCEPTED_KWARGS]
    if wrong_kwargs:
        raise KeyError(f'Unknown kwarg(s): {",".join(wrong_kwargs)}. '
                       f'Currently accepted: {",".join(_ACCEPTED_KWARGS)}')

    _METHOD_OPTIONS = ['2d', '3d', '3d_complex']
    if method not in _METHOD_OPTIONS:
        raise ValueError(f'Unknown method "{method}". Please use either: '
                         f'{",".join(_METHOD_OPTIONS)}')

    # Set axis to plot for method '2d'
    axis1, axis2 = kwargs.get('view', ('x', 'y'))

    plot_soma = kwargs.get('soma', True)

    connectors = kwargs.get('connectors', False)
    connectors_only = kwargs.get('connectors_only', False)
    cn_mesh_colors = kwargs.get('cn_mesh_colors', False)
    use_neuron_color = kwargs.get('use_neuron_color', False)
    ax = kwargs.get('ax', None)
    color = kwargs.get('color',
                       kwargs.get('c',
                                  kwargs.get('colors', None)))
    scalebar = kwargs.get('scalebar', None)
    group_neurons = kwargs.get('group_neurons', False)

    # This is overwritten if color specifies alphas
    alpha = kwargs.get('alpha', .9)

    # Depth coloring
    depth_coloring = kwargs.get('depth_coloring', False)
    depth_scale = kwargs.get('depth_scale', True)

    scatter_kws = kwargs.get('scatter_kws', {})

    linewidth = kwargs.get('linewidth', kwargs.get('lw', .5))
    cn_size = kwargs.get('cn_size', 1)
    linestyle = kwargs.get('linestyle', kwargs.get('ls', '-'))
    autoscale = kwargs.get('autoscale', True)

    # Keep track of limits if necessary
    lim = []

    # Parse objects
    skdata, dotprops, volumes, points, visuals = utils.parse_objects(x)

    # Generate the colormaps
    (neuron_cmap,
     dotprop_cmap,
     volumes_cmap) = prepare_colormap(color, skdata, dotprops, volumes,
                                      use_neuron_color=use_neuron_color,
                                      color_range=1)

    # Make sure axes are projected orthogonally
    if method in ['3d', '3d_complex']:
        proj3d.persp_transformation = _orthogonal_proj

    # Generate axes
    if not ax:
        if method == '2d':
            fig, ax = plt.subplots(figsize=kwargs.get('figsize', (8, 8)))
            ax.set_aspect('equal')
        elif method in ['3d', '3d_complex']:
            fig = plt.figure(figsize=kwargs.get('figsize',
                                                plt.figaspect(1) * 1.5))
            ax = fig.gca(projection='3d')

            # This sets front view
            ax.azim = -90
            ax.elev = 0
            ax.dist = 7
            # Disallowed for 3D in matplotlib 3.1.0
            # ax.set_aspect('equal')
    # Check if correct axis were provided
    else:
        if not isinstance(ax, mpl.axes.Axes):
            raise TypeError('Ax must be of type "mpl.axes.Axes", '
                            f'not "{type(ax)}"')
        fig = ax.get_figure()
        if method in ['3d', '3d_complex']:
            if ax.name != '3d':
                raise TypeError('Axis must be 3d.')

            # Add existing limits to ax
            lim += np.array((ax.get_xlim3d(),
                             ax.get_zlim3d(),
                             ax.get_ylim3d())
                            ).T.tolist()
        elif method == '2d':
            if ax.name == '3d':
                raise TypeError('Axis must be 2d.')

    # Kep track of whether ax already has data
    ax_had_data = ax.has_data()

    # Prepare some stuff for depth coloring
    if depth_coloring and method == '3d_complex':
        raise Exception(f'Depth coloring unavailable for method "{method}"')
    elif depth_coloring and method == '2d':
        all_co = skdata.nodes[['x', 'y', 'z']]
        norm = plt.Normalize(vmin=all_co.z.min(), vmax=all_co.z.max())

    # Plot volumes first
    if volumes:
        for i, v in enumerate(volumes):
            c = volumes_cmap[i]

            if len(c) == 4:
                this_alpha = c[3]
            else:
                this_alpha = 1

            if kwargs.get('volume_outlines', False):
                fill, lw, fc, ec = False, 1, 'none', c
            else:
                fill, lw, fc, ec = True, 0, c, 'none'

            if method == '2d':
                verts = v.to_2d(view=f'{axis1}{axis2}', invert_y=True)
                vpatch = mpatches.Polygon(verts, closed=True, lw=lw, fill=fill,
                                          fc=fc, ec=ec, alpha=this_alpha, zorder=0)
                ax.add_patch(vpatch)
            elif method in ['3d', '3d_complex']:
                verts = np.vstack(v.vertices)

                # Add alpha
                if len(c) == 3:
                    c = (c[0], c[1], c[2], .1)

                ts = ax.plot_trisurf(verts[:, 0],
                                     verts[:, 1],
                                     v.faces,
                                     verts[:, 2],
                                     label=v.name,
                                     color=c)
                ts.set_gid(v.name)

                ax_had_data = _update_axes3d_bounds(ax,
                                                    verts[:, [0, 1, 2]],
                                                    had_data=ax_had_data)

    # Create lines from segments
    line3D_collections = []
    surf3D_collections = []
    for i, neuron in enumerate(config.tqdm(skdata.itertuples(),
                                           desc='Plot neurons',
                                           total=skdata.shape[0],
                                           leave=False,
                                           disable=config.pbar_hide | len(dotprops) == 0)):
        this_color = neuron_cmap[i]

        if neuron.nodes.empty:
            logger.warning(f'Skipping neuron w/o nodes: {neuron.id}')
            continue

        if not connectors_only:
            # Now make traces (invert y axis)
            coords = segments_to_coords(neuron,
                                        neuron.segments,
                                        modifier=(1, 1, 1))

            if method == '2d':
                if not depth_coloring:
                    # We have to add (None, None, None) to the end of each
                    # slab to make that line discontinuous there
                    coords = np.vstack(
                        [np.append(t, [[None] * 3], axis=0) for t in coords])

                    this_line = mlines.Line2D(coords[:, 0], coords[:, 1],
                                              lw=linewidth, ls=linestyle,
                                              alpha=alpha, color=this_color,
                                              label=f'{getattr(neuron, "name", "NA")} - #{neuron.id}')
                    ax.add_line(this_line)
                else:
                    coords = tn_pairs_to_coords(neuron, modifier=(1, 1, 1))
                    lc = LineCollection(coords[:, :, [0, 1]],
                                        cmap='jet',
                                        norm=norm)
                    lc.set_array(neuron.nodes.loc[neuron.nodes.parent_id >= 0, 'z'].values)
                    lc.set_linewidth(linewidth)
                    lc.set_alpha(alpha)
                    lc.set_linestyle(linestyle)
                    lc.set_label(f'{getattr(neuron, "name", "NA")} - #{neuron.id}')
                    line = ax.add_collection(lc)

                if plot_soma and not isinstance(neuron.soma, type(None)):
                    soma = utils.make_iterable(neuron.soma)
                    for s in soma:
                        n = neuron.nodes.set_index('node_id').loc[s]
                        r = getattr(n, neuron.soma_radius) if isinstance(neuron.soma_radius, str) else neuron.soma_radius

                        if depth_coloring:
                            this_color = mpl.cm.jet(norm(n.z))

                        s = mpatches.Circle((int(n.x), int(n.y)), radius=r,
                                            alpha=alpha, fill=True, fc=this_color,
                                            zorder=4, edgecolor='none')
                        ax.add_patch(s)

            elif method in ['3d', '3d_complex']:
                cmap = mpl.cm.jet if depth_coloring else None

                # For simple scenes, add whole neurons at a time -> will speed
                # up rendering
                if method == '3d':
                    if depth_coloring:
                        this_coords = tn_pairs_to_coords(neuron,
                                                         modifier=(1, 1, 1))[:, :, [0, 1, 2]]
                    else:
                        this_coords = [c[:, [0, 1, 2]] for c in coords]

                    lc = Line3DCollection(this_coords,
                                          color=this_color,
                                          label=neuron.id,
                                          alpha=alpha,
                                          cmap=cmap,
                                          lw=linewidth,
                                          linestyle=linestyle)
                    if group_neurons:
                        lc.set_gid(neuron.id)
                    # Need to get this before adding data
                    ax.add_collection3d(lc)
                    # Update data bounds
                    ax_had_data = _update_axes3d_bounds(ax,
                                                        neuron.nodes[['x', 'y', 'z']].values,
                                                        had_data=ax_had_data)
                    line3D_collections.append(lc)

                # For complex scenes, add each segment as a single collection
                # -> helps preventing Z-order errors
                elif method == '3d_complex':
                    for c in coords:
                        lc = Line3DCollection([c],
                                              color=this_color,
                                              lw=linewidth,
                                              alpha=alpha,
                                              linestyle=linestyle)
                        if group_neurons:
                            lc.set_gid(neuron.id)
                        ax.add_collection3d(lc)
                    ax_had_data = _update_axes3d_bounds(ax,
                                                        neuron.nodes[['x', 'y', 'z']].values,
                                                        had_data=ax_had_data)

                surf3D_collections.append([])
                if plot_soma and not isinstance(neuron.soma, type(None)):
                    soma = utils.make_iterable(neuron.soma)
                    for s in soma:
                        n = neuron.nodes.set_index('node_id').loc[s]
                        r = getattr(n, neuron.soma_radius) if isinstance(neuron.soma_radius, str) else neuron.soma_radius

                        resolution = 20
                        u = np.linspace(0, 2 * np.pi, resolution)
                        v = np.linspace(0, np.pi, resolution)
                        x = r * np.outer(np.cos(u), np.sin(v)) + n.x
                        y = r * np.outer(np.sin(u), np.sin(v)) + n.y
                        z = r * np.outer(np.ones(np.size(u)), np.cos(v)) + n.z
                        surf = ax.plot_surface(x, y, z,
                                               color=this_color,
                                               shade=False,
                                               alpha=alpha)
                        if group_neurons:
                            surf.set_gid(neuron.id)

                        surf3D_collections[-1].append(surf)

        if (connectors or connectors_only) and neuron.has_connectors:
            if not cn_mesh_colors:
                cn_lay = config.default_connector_colors
            else:
                cn_lay = {{'name': c, 'color': this_color}
                          for c in neuron.connectors.type.unique()}
            if method == '2d':
                for c in neuron.connectors.type.unique():
                    this_cn = neuron.connectors[neuron.connectors.type == c]
                    ax.scatter(this_cn.x.values,
                               this_cn.y.values,
                               c=cn_lay[c]['color'], alpha=alpha, zorder=4,
                               edgecolor='none', s=cn_size)
                    ax.get_children()[-1].set_gid(f'CN_{neuron.id}')
            elif method in ['3d', '3d_complex']:
                all_cn = neuron.connectors
                c = [cn_lay[i]['color'] for i in all_cn.type.values]
                ax.scatter(all_cn.x.values, all_cn.y.values, all_cn.z.values,
                           c=c, s=cn_size, depthshade=False, edgecolor='none',
                           alpha=alpha)
                ax.get_children()[-1].set_gid(f'CN_{neuron.id}')
                # Update data bounds
                ax_had_data = _update_axes3d_bounds(ax,
                                                    all_cn[['x', 'y', 'z']].values,
                                                    had_data=ax_had_data)

    for i, neuron in enumerate(config.tqdm(dotprops.itertuples(),
                                           desc='Plt dotprops',
                                           total=dotprops.shape[0],
                                           leave=False,
                                           disable=config.pbar_hide | len(dotprops) == 0)):
        # Prepare lines - this is based on nat:::plot3d.dotprops
        halfvect = neuron.points[['x_vec', 'y_vec', 'z_vec']] / 2

        starts = neuron.points[['x', 'y', 'z']].values - halfvect.values
        ends = neuron.points[['x', 'y', 'z']].values + halfvect.values

        try:
            this_color = dotprop_cmap[i]
        except BaseException:
            this_color = (.1, .1, .1)

        if method == '2d':
            # Add None between segments
            x_coords = [n for sublist in zip(starts[:, 0],
                                             ends[:, 0],
                                             [None] * starts.shape[0]) for n in sublist]
            y_coords = [n for sublist in zip(starts[:, 1],
                                             ends[:, 1],
                                             [None] * starts.shape[0]) for n in sublist]

            this_line = mlines.Line2D(x_coords, y_coords,
                                      lw=linewidth, ls=linestyle,
                                      alpha=alpha, color=this_color,
                                      label='%s' % (neuron.gene_name))

            ax.add_line(this_line)

            # Add soma
            if plot_soma:
                s = mpatches.Circle((neuron.X, neuron.Y), radius=2,
                                    alpha=alpha, fill=True, fc=this_color,
                                    zorder=4, edgecolor='none')
                ax.add_patch(s)
        elif method in ['3d', '3d_complex']:
            # Combine coords by weaving starts and ends together
            coords = np.empty((starts.shape[0] * 2, 3), dtype=starts.dtype)
            coords[0::2] = starts
            coords[1::2] = ends

            # Invert y-axis
            # coords[:, 1] *= -1

            # For simple scenes, add whole neurons at a time
            # -> will speed up rendering
            if method == '3d':
                lc = Line3DCollection(np.split(coords,
                                               starts.shape[0]),
                                      color=this_color,
                                      label=neuron.gene_name,
                                      lw=linewidth,
                                      alpha=alpha,
                                      linestyle=linestyle)
                if group_neurons:
                    lc.set_gid(neuron.gene_name)
                ax.add_collection3d(lc)

            # For complex scenes, add each segment as a single collection
            # -> help preventing Z-order errors
            elif method == '3d_complex':
                for c in np.split(coords, starts.shape[0]):
                    lc = Line3DCollection([c],
                                          color=this_color,
                                          lw=linewidth,
                                          alpha=alpha,
                                          linestyle=linestyle)
                    if group_neurons:
                        lc.set_gid(neuron.gene_name)
                    ax.add_collection3d(lc)

            ax_had_data = _update_axes3d_bounds(ax,
                                                coords,
                                                had_data=ax_had_data)

            lim.append(coords.max(axis=0))
            lim.append(coords.min(axis=0))

            resolution = 20
            u = np.linspace(0, 2 * np.pi, resolution)
            v = np.linspace(0, np.pi, resolution)
            x = 2 * np.outer(np.cos(u), np.sin(v)) + neuron.X
            y = 2 * np.outer(np.sin(u), np.sin(v)) + neuron.Y
            z = 2 * np.outer(np.ones(np.size(u)), np.cos(v)) + neuron.Z
            surf = ax.plot_surface(x, y, z,
                                   color=this_color,
                                   shade=False,
                                   alpha=alpha)
            if group_neurons:
                surf.set_gid(neuron.gene_name)

    if points:
        for p in points:
            if method == '2d':
                default_settings = dict(
                    c='black',
                    zorder=4,
                    edgecolor='none',
                    s=1
                )
                default_settings.update(scatter_kws)
                default_settings = _fix_default_dict(default_settings)

                ax.scatter(p[:, 0],
                           p[:, 1],
                           **default_settings)
            elif method in ['3d', '3d_complex']:
                default_settings = dict(
                    c='black',
                    s=1,
                    depthshade=False,
                    edgecolor='none'
                )
                default_settings.update(scatter_kws)
                default_settings = _fix_default_dict(default_settings)

                ax.scatter(p[:, 0], p[:, 1], p[:, 2],
                           **default_settings
                           )
                ax_had_data = _update_axes3d_bounds(ax,
                                                    p[:, [0, 1, 2]],
                                                    had_data=ax_had_data)

    if autoscale:
        if method == '2d':
            ax.autoscale(tight=True)
        elif method in ['3d', '3d_complex']:
            # First autoscale
            ax.autoscale()
            # Now we need to set equal aspect manually
            lim = np.array([ax.get_xlim(),
                            ax.get_ylim(),
                            ax.get_zlim()])
            dim = lim[:, 1] - lim[:, 0]
            center = lim[:, 0] + dim / 2
            max_dim = dim.max()

            new_min = center - max_dim / 2
            new_max = center + max_dim / 2

            ax.set_xlim(new_min[0], new_max[0])
            ax.set_ylim(new_min[1], new_max[1])
            ax.set_zlim(new_min[2], new_max[2])

    if scalebar is not None:
        if isinstance(scalebar, bool):
            scalebar = '1 um'

        if isinstance(scalebar, str):
            scalebar = config.ureg(scalebar)

        if isinstance(scalebar, pint.Quantity):
            # If we have neurons as points of reference convert
            if skdata:
                scalebar = scalebar.to(skdata[0].units).magnitude
            # If no reference, use assume it's the same units
            else:
                scalebar = scalebar.magnitude

        # Hard-coded offset from figure boundaries
        ax_offset = (ax.get_xlim()[1] - ax.get_xlim()[0]) / 100 * 5

        if method == '2d':
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()

            coords = np.array([[xlim[0] + ax_offset, ylim[0] + ax_offset],
                               [xlim[0] + ax_offset + scalebar, ylim[0] + ax_offset]
                               ])

            print(coords)

            sbar = mlines.Line2D(
                coords[:, 0], coords[:, 1], lw=3, alpha=.9, color='black')
            sbar.set_gid(f'{scalebar}_scalebar')

            ax.add_line(sbar)
        elif method in ['3d', '3d_complex']:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
            zlim = ax.get_zlim()

            left = xlim[0] + ax_offset
            bottom = zlim[0] + ax_offset
            front = ylim[0] + ax_offset

            sbar = [np.array([[left, front, bottom],
                              [left, front, bottom]]),
                    np.array([[left, front, bottom],
                              [left, front, bottom]]),
                    np.array([[left, front, bottom],
                              [left, front, bottom]])]
            sbar[0][1][0] += scalebar
            sbar[1][1][1] += scalebar
            sbar[2][1][2] += scalebar

            lc = Line3DCollection(sbar, color='black', lw=1)
            lc.set_gid(f'{scalebar}_scalebar')
            ax.add_collection3d(lc)

    def set_depth():
        """Sets depth information for neurons according to camera position."""

        # Modifier for soma coordinates
        modifier = np.array([1, 1, 1])

        # Get all coordinates
        all_co = np.concatenate([lc._segments3d[:, 0, :] for lc in line3D_collections],
                                axis=0)

        # Get projected coordinates
        proj_co = mpl_toolkits.mplot3d.proj3d.proj_points(all_co, ax.get_proj())

        # Get min and max of z coordinates
        z_min, z_max = min(proj_co[:, 2]), max(proj_co[:, 2])

        # Generate a new normaliser
        norm = plt.Normalize(vmin=z_min, vmax=z_max)

        # Go over all neurons and update Z information
        for neuron, lc, surf in zip(skdata,
                                    line3D_collections,
                                    surf3D_collections):
            # Get this neurons coordinates
            this_co = lc._segments3d[:, 0, :]

            # Get projected coordinates
            this_proj = mpl_toolkits.mplot3d.proj3d.proj_points(this_co,
                                                                ax.get_proj())

            # Normalise z coordinates
            ns = norm(this_proj[:, 2]).data

            # Set array
            lc.set_array(ns)

            # No need for normaliser - already happened
            lc.set_norm(None)

            if not isinstance(neuron.soma, type(None)):
                # Get depth of soma(s)
                soma = utils.make_iterable(neuron.soma)
                soma_co = neuron.nodes.set_index('node_id').loc[soma][['x', 'y', 'z']].values
                soma_proj = mpl_toolkits.mplot3d.proj3d.proj_points(soma_co * modifier,
                                                                    ax.get_proj())
                soma_cs = norm(soma_proj[:, 2]).data

                # Set soma color
                for cs, s in zip(soma_cs, surf):
                    s.set_color(cmap(cs))

    def Update(event):
        set_depth()

    if depth_coloring:
        if method == '2d' and depth_scale:
            fig.colorbar(line, ax=ax, fraction=.075, shrink=.5, label='Depth')
        elif method == '3d':
            fig.canvas.mpl_connect('draw_event', Update)
            set_depth()

    plt.axis('off')

    logger.debug('Done. Use matplotlib.pyplot.show() to show plot.')

    return fig, ax


def _update_axes3d_bounds(ax, points, had_data=False):
    """Update axis bounds and remove default points (0,0,0) and (1,1,1)."""
    if not isinstance(points, np.ndarray):
        points = np.ndarray(points)

    # If this is the first set of points, we need to overwrite the defaults
    # That should happen automatically but for some reason doesn't for 3d axes
    if not had_data:
        mn = points.min(axis=0)
        mx = points.max(axis=0)
        new_xybounds = np.array([[mn[0], mn[1]],
                                 [mx[0], mx[1]]])
        new_zzbounds = np.array([[mn[2], mn[2]],
                                 [mx[2], mx[2]]])
        ax.xy_dataLim.set_points(new_xybounds)
        ax.zz_dataLim.set_points(new_zzbounds)
    else:
        ax.auto_scale_xyz(points[:, 0].tolist(),
                          points[:, 1].tolist(),
                          points[:, 2].tolist(),
                          had_data=True)

    return True


def _fix_default_dict(x):
    """ Consolidates duplicate settings in e.g. scatter kwargs when 'c' and
    'color' is provided.
    """

    # The first entry is the "survivor"
    duplicates = [['color', 'c'], ['size', 's'], ['alpha', 'a']]

    for dupl in duplicates:
        if sum([v in x for v in dupl]) > 1:
            to_delete = [v for v in dupl if v in x][1:]
            _ = [x.pop(v) for v in to_delete]

    return x


def _orthogonal_proj(zfront, zback):
    """ Function to get matplotlib to use orthogonal instead of perspective
    view.

    Usage:
    proj3d.persp_transformation = _orthogonal_proj
    """
    a = (zfront + zback) / (zfront - zback)
    b = -2 * (zfront * zback) / (zfront - zback)
    # -0.0001 added for numerical stability as suggested in:
    # http://stackoverflow.com/questions/23840756
    return np.array([[1, 0, 0, 0],
                     [0, 1, 0, 0],
                     [0, 0, a, b],
                     [0, 0, -0.0001, zback]])



#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Set of functions to interface with the www.insectbraindb.org database of
insect brains and neurons.
"""

import os
import requests
import tempfile

import pandas as pd

from tqdm import tqdm

from typing import Union, Optional, List

from ..core import Volume
from ..utils import make_url
from ..config import logger
from ..plotting.colors import hex_to_rgb


baseurl = 'http://www.insectbraindb.org'


def get_brain_meshes(species: Union[str, int],
                     save_to: Optional[str] = None,
                     combine_vols: bool = False
                     ) -> Optional[List[Volume]]:
    """ Fetch brain meshes for given species.given

    Parameters
    ----------
    species:        str | int
                    Species for which to fetch brain volumes. Strings are
                    interpreted as names, integers as IDs.
    save_to:        str, optional
                    Folder to save .obj files to. If provided, will NOT return
                    navis.Volume.
    combine_vols :  bool, optional
                    If True, will combine subvolumes (i.e. neuropils) into
                    a single navis.Volume - else will return list with volumes.

    Returns
    -------
    None
                    If ``save_to`` is folder.
    list of navis.Volume
                    If ``combine_vols`` is False.

    Examples
    --------
    >>> import navis
    >>> import navis.interfaces.insectbrain_db as ibd
    >>> v = ibd.get_brain_meshes('Desert Locust', combine_vols=True)
    >>> navis.plot3d(v)

    """

    if isinstance(save_to, str):
        if not os.path.isdir(save_to):
            raise ValueError('"save_to" must be None or valid path.')

    obj_url = 'https://s3.eu-central-1.amazonaws.com/ibdb-file-storage/'

    # Get info with all available neuropils
    sp_info = get_species_info(species)

    # Go over all brains
    n_brains = len(sp_info.reconstructions)  # type: ignore
    n_reconstr = len([r for r in sp_info.reconstructions if r.get('viewer_files')])  # type: ignore
    logger.info(f'{n_reconstr} reconstruction(s) from {n_brains} brain(s) found')

    volumes: List[Volume] = []
    for brain in tqdm(sp_info.reconstructions, desc='Brains'):  # type: ignore
        this_v = []
        # If no reconstructions, continue
        if not brain.get('viewer_files'):  # type: ignore
            continue
        for file in tqdm(brain['viewer_files'], desc='Meshes'):
            filename = file['p_file']['file_name']
            path = file['p_file']['path']
            url = make_url(obj_url, path)

            if file.get('neuropil'):
                color = hex_to_rgb(file['neuropil'][0]['color'])
            else:
                color = (1, 1, 1, .1)

            resp = requests.get(url)
            resp.raise_for_status()

            if save_to:
                fp = os.path.join(save_to, filename)
            else:
                # Save temporary file
                fp = os.path.join(tempfile.gettempdir(), 'temp.obj')

            with open(fp, 'w') as f:
                f.write(resp.content.decode())

            if not save_to:
                v = Volume.from_file(fp, name=filename, color=color)
                this_v.append(v)

        # Combine all volumes in this brain
        if combine_vols:
            this_v = [Volume.combine(this_v)]

        volumes += this_v

    if not save_to:
        return volumes
    else:
        return None


def get_species_info(species: Union[str, int]) -> pd.Series:
    """ Get all info for given species.

    Parameters
    ----------
    species :       str | int
                    Species to get info for.

    Returns
    -------
    pandas.Series
                    Pandas Series with info on given species.

    Examples
    --------
    >>> import navis.interfaces.insectbrain_db as ibd
    >>> info = ibd.get_species_info()
    """

    # First get species ID
    if not isinstance(species, int):
        all_species = get_available_species()
        if species in all_species.common_name.values:
            sid = all_species.set_index('common_name',
                                        inplace=False).loc[species, 'id']
        elif species in all_species.scientific_name.values:
            sid = all_species.set_index('scientific_name',
                                        inplace=False).loc[species, 'id']
        else:
            raise ValueError(f'Species "{species}" not found.')
    else:
        sid = species

    url = make_url(baseurl, '/archive/species/most_current_permitted/',
                   species_id=sid)

    resp = requests.get(url)

    resp.raise_for_status()

    return pd.Series(resp.json())


def get_available_species() -> pd.DataFrame:
    """ Get all info for given species.

    Examples
    --------
    >>> import navis.interfaces.insectbrain_db as ibd
    >>> specs = ibd.get_available_species()

    Returns
    -------
    pandas.DataFrame
            DataFrame with available species.

    Examples
    --------
    >>> import navis.interfaces.insectbrain_db as ibd
    >>> species = ibd.get_available_species()
    """

    url = make_url(baseurl, 'api', 'species', 'min')

    resp = requests.get(url)

    resp.raise_for_status()

    return pd.DataFrame.from_records(resp.json())



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


""" This module contains functions to analyse and manipulate neuron morphology.
"""

import itertools

import pandas as pd
import numpy as np
import scipy.spatial.distance
import networkx as nx

from typing import Union, Optional, Sequence, overload, List, Set
from typing_extensions import Literal

from .. import graph, utils, config, core
from . import mmetrics

# Set up logging
logger = config.logger

__all__ = sorted(['prune_by_strahler', 'stitch_neurons', 'split_axon_dendrite',
                  'average_neurons', 'despike_neuron', 'guess_radius',
                  'smooth_neuron', 'heal_fragmented_neuron',
                  'break_fragments', 'prune_twigs'])

NeuronObject = Union['core.NeuronList', 'core.TreeNeuron']


@overload
def prune_by_strahler(x: NeuronObject,
                      to_prune: Union[int, List[int], range, slice],
                      inplace: Literal[False],
                      reroot_soma: bool = True,
                      force_strahler_update: bool = False,
                      relocate_connectors: bool = False) -> NeuronObject: ...


@overload
def prune_by_strahler(x: NeuronObject,
                      to_prune: Union[int, List[int], range, slice],
                      inplace: Literal[True],
                      reroot_soma: bool = True,
                      force_strahler_update: bool = False,
                      relocate_connectors: bool = False) -> None: ...


@overload
def prune_by_strahler(x: NeuronObject,
                      to_prune: Union[int, List[int], range, slice],
                      inplace: bool = False,
                      reroot_soma: bool = True,
                      force_strahler_update: bool = False,
                      relocate_connectors: bool = False) -> Optional[NeuronObject]: ...


def prune_by_strahler(x: NeuronObject,
                      to_prune: Union[int, List[int], range, slice],
                      inplace: bool = False,
                      reroot_soma: bool = True,
                      force_strahler_update: bool = False,
                      relocate_connectors: bool = False) -> Optional[NeuronObject]:
    """ Prune neuron based on `Strahler order
    <https://en.wikipedia.org/wiki/Strahler_number>`_.

    Parameters
    ----------
    x :             TreeNeuron | NeuronList
    to_prune :      int | list | range | slice
                    Strahler indices (SI) to prune. For example:

                    1. ``to_prune=1`` removes all leaf branches
                    2. ``to_prune=[1, 2]`` removes SI 1 and 2
                    3. ``to_prune=range(1, 4)`` removes SI 1, 2 and 3
                    4. ``to_prune=slice(0, -1)`` removes everything but the
                       highest SI
                    5. ``to_prune=slice(-1, None)`` removes only the highest
                       SI

    reroot_soma :   bool, optional
                    If True, neuron will be rerooted to its soma.
    inplace :       bool, optional
                    If False, pruning is performed on copy of original neuron
                    which is then returned.
    force_strahler_update : bool, optional
                            If True, will force update of Strahler order even
                            if already exists in node table.
    relocate_connectors : bool, optional
                          If True, connectors on removed nodes will be
                          reconnected to the closest still existing node.
                          Works only in child->parent direction.

    Returns
    -------
    TreeNeuron/List
                    Pruned neuron(s).

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> n_pr = navis.prune_by_strahler(n, to_prune=1, inplace=False)
    >>> n.n_nodes > n_pr.n_nodes
    True

    """

    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()

        [prune_by_strahler(n,
                           to_prune=to_prune,
                           inplace=True,
                           reroot_soma=reroot_soma,
                           force_strahler_update=force_strahler_update,
                           relocate_connectors=relocate_connectors
                           ) for n in x]
        if not inplace:
            return x
        else:
            return None
    elif isinstance(x, core.TreeNeuron):
        neuron = x
    else:
        raise TypeError(f'Expected Neuron/List, got {type(x)}')

    # Make a copy if necessary before making any changes
    if not inplace:
        neuron = neuron.copy()

    if reroot_soma and neuron.soma:
        neuron.reroot(neuron.soma, inplace=True)

    if 'strahler_index' not in neuron.nodes or force_strahler_update:
        mmetrics.strahler_index(neuron)

    # Prepare indices
    if isinstance(to_prune, int) and to_prune < 0:
        to_prune = range(1, int(neuron.nodes.strahler_index.max() + (to_prune + 1)))

    if isinstance(to_prune, int):
        if to_prune < 1:
            raise ValueError('SI to prune must be positive. Please see help'
                             'for additional options.')
        to_prune = [to_prune]
    elif isinstance(to_prune, range):
        to_prune = list(to_prune)
    elif isinstance(to_prune, slice):
        SI_range = range(1, int(neuron.nodes.strahler_index.max() + 1))
        to_prune = list(SI_range)[to_prune]

    # Prepare parent dict if needed later
    if relocate_connectors:
        parent_dict = {
            tn.node_id: tn.parent_id for tn in neuron.nodes.itertuples()}

    neuron.nodes = neuron.nodes[~neuron.nodes.strahler_index.isin(to_prune)].reset_index(drop=True, inplace=False)

    if neuron.has_connectors:
        if not relocate_connectors:
            neuron.connectors = neuron.connectors[neuron.connectors.node_id.isin(neuron.nodes.node_id.values)].reset_index(drop=True, inplace=False)
        else:
            remaining_tns = set(neuron.nodes.node_id.values)
            for cn in neuron.connectors[~neuron.connectors.node_id.isin(neuron.nodes.node_id.values)].itertuples():
                this_tn = parent_dict[cn.node_id]
                while True:
                    if this_tn in remaining_tns:
                        break
                    this_tn = parent_dict[this_tn]
                neuron.connectors.loc[cn.Index, 'node_id'] = this_tn

    # Reset indices of node and connector tables (important for igraph!)
    neuron.nodes.reset_index(inplace=True, drop=True)

    if neuron.has_connectors:
        neuron.connectors.reset_index(inplace=True, drop=True)

    # Theoretically we can end up with disconnected pieces, i.e. with more
    # than 1 root node -> we have to fix the nodes that lost their parents
    neuron.nodes.loc[~neuron.nodes.parent_id.isin(neuron.nodes.node_id.values), 'parent_id'] = -1

    # Remove temporary attributes
    neuron._clear_temp_attr()

    if not inplace:
        return neuron
    else:
        return None


@overload
def prune_twigs(x: NeuronObject,
                size: float,
                inplace: Literal[True],
                recursive: Union[int, bool, float] = False
                ) -> None: ...


@overload
def prune_twigs(x: NeuronObject,
                size: float,
                inplace: Literal[False],
                recursive: Union[int, bool, float] = False
                ) -> NeuronObject: ...


@overload
def prune_twigs(x: NeuronObject,
                size: float,
                inplace: bool = False,
                recursive: Union[int, bool, float] = False
                ) -> Optional[NeuronObject]: ...


def prune_twigs(x: NeuronObject,
                size: float,
                inplace: bool = False,
                recursive: Union[int, bool, float] = False
                ) -> Optional[NeuronObject]:
    """ Prune terminal twigs under a given size.

    Parameters
    ----------
    x :             TreeNeuron | NeuronList
    size :          int | float
                    Twigs shorter than this will be pruned.
    inplace :       bool, optional
                    If False, pruning is performed on copy of original neuron
                    which is then returned.
    recursive :     int | bool | float("inf"), optional
                    If `int` will undergo that many rounds of recursive
                    pruning. Use `float(`"inf")`` to prune until no more twigs
                    under the given size are left.

    Returns
    -------
    TreeNeuron/List
                    Pruned neuron(s).

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> # Prune twigs smaller than 5 microns (example neurons are in nm)
    >>> n_pr = navis.prune_twigs(n,
    ...                          size=5000,
    ...                          recursive=float('inf'),
    ...                          inplace=False)
    >>> n.n_nodes > n_pr.n_nodes
    True

    """

    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()

        [prune_twigs(n,
                     size=size,
                     inplace=True,
                     recursive=recursive) for n in tqdm(x,
                                                        desc='Pruning',
                                                        disable=config.pbar_hide,
                                                        leave=config.pbar_leave)]

        if not inplace:
            return x
        else:
            return None
    elif isinstance(x, core.TreeNeuron):
        neuron = x
    else:
        raise TypeError(f'Expected Neuron/List, got {type(x)}')

    # If people set recursive=True, assume that they mean float("inf")
    if isinstance(recursive, bool) and recursive:
        recursive = float('inf')

    # Make a copy if necessary before making any changes
    if not inplace:
        neuron = neuron.copy()

    # Find terminal nodes
    leafs = neuron.nodes[neuron.nodes.type == 'end'].node_id.values

    # Find terminal segments
    segs = graph._break_segments(neuron)
    segs = np.array([s for s in segs if s[0] in leafs])

    # Get segment lengths
    seg_lengths = np.array([graph.segment_length(neuron, s) for s in segs])

    # Find out which to delete
    segs_to_delete = segs[seg_lengths <= size]

    if segs_to_delete.any():
        # Unravel the into list of node IDs -> skip the last parent
        nodes_to_delete = [n for s in segs_to_delete for n in s[:-1]]

        # Subset neuron
        nodes_to_keep = neuron.nodes[~neuron.nodes.node_id.isin(nodes_to_delete)].node_id.values
        graph.subset_neuron(neuron,
                            nodes_to_keep,
                            inplace=True)

        # Go recursive
        if recursive:
            recursive -= 1
            prune_twigs(neuron, size=size, inplace=True, recursive=recursive)

    if not inplace:
        return neuron
    else:
        return None


def split_axon_dendrite(x: NeuronObject,
                        method: Union[Literal['centrifugal'],
                                      Literal['centripetal'],
                                      Literal['sum'],
                                      Literal['bending']] = 'bending',
                        split: Union[Literal['prepost'],
                                     Literal['distance']] = 'prepost',
                        reroot_soma: bool = True,
                        return_point: bool = False) -> 'core.NeuronList':
    """Split a neuron into axon, dendrite and primary neurite.

    The result is highly dependent on the method and on your neuron's
    morphology and works best for "typical" neurons, i.e. those where the
    primary neurite branches into axon and dendrites.

    Parameters
    ----------
    x :                 TreeNeuron | NeuronList
                        Neuron(s) to split into axon, dendrite (and primary
                        neurite if possible). MUST HAVE CONNECTORS.
    method :            'centrifugal' | 'centripetal' | 'sum' | 'bending', optional
                        Type of flow centrality to use to split the neuron.
                        There are four flavors: the first three refer to
                        :func:`~navis.flow_centrality`, the last
                        refers to :func:`~navis.bending_flow`.

                        Will try using stored centrality, if possible.
    split :             "prepost" | "distance"
                        Method for determining which compartment is axon and
                        which is the dendrites:

                            - "prepost" uses number of in- vs. outputs
                            - "distance" assumes the compartment closer to the
                              soma is the dendrites


    reroot_soma :       bool, optional
                        If True, will make sure neuron is rooted to soma if at
                        all possible.
    return_point :      bool, optional
                        If True, will only return node ID of the node at
                        which to split the neuron.

    Returns
    -------
    NeuronList
                        Axon, dendrite and primary neurite. Fragments will
                        have a new property ``compartment`` (see example).

    Examples
    --------
    >>> import navis
    >>> x = navis.example_neurons(1)
    >>> split = navis.split_axon_dendrite(x, method='centrifugal',
    ...                                   reroot_soma=True)
    >>> split
    <class 'navis.NeuronList'> of 3 neurons
                          neuron_name skeleton_id  n_nodes  n_connectors
    0                  neuron 123457          16      148             0
    1                  neuron 123457          16     9682          1766
    2                  neuron 123457          16     2892           113
    >>> # For convenience, split_axon_dendrite assigns colors to the resulting
    >>> # fragments: axon = red, dendrites = blue, primary neurite = green
    >>> split.plot3d(color=split.color)

    See Also
    --------
    :func:`navis.heal_fragmented_neuron`
            Axon/dendrite split works only on neurons consisting of a single
            tree. Use this function to heal fragmented neurons.

    """
    if isinstance(x, core.NeuronList) and len(x) == 1:
        x = x[0]
    elif isinstance(x, core.NeuronList):
        nl = []
        for n in config.tqdm(x, desc='Splitting', disable=config.pbar_hide,
                             leave=config.pbar_leave):
            nl.append(split_axon_dendrite(n,
                                          method=method,
                                          split=split,
                                          reroot_soma=reroot_soma,
                                          return_point=return_point))
        return core.NeuronList([n for l in nl for n in l])

    if not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Can only process TreeNeuron, got "{type(x)}"')

    if not x.has_connectors:
        raise ValueError('Neuron must have connectors.')

    if method not in ['centrifugal', 'centripetal', 'sum', 'bending']:
        raise ValueError(f'"{method}" not allowed for parameter `method`.')

    if split not in ['prepost', 'distance']:
        raise ValueError(f'"{split}" not allowed for parameter `split`.')

    if len(x.root) > 1:
        raise ValueError(f'Unable to split neuron {x.id}: multiple roots. '
                         'Try navis.heal_fragmented_neuron(x) to merged '
                         'disconnected fragments.')

    if x.soma and x.soma not in x.root and reroot_soma:
        x.reroot(x.soma, inplace=True)

    # Calculate flow centrality if necessary
    last_method = getattr(x, 'centrality_method', None)

    if last_method != method:
        if method == 'bending':
            mmetrics.bending_flow(x)
        elif method in ['centripetal', 'centrifugal', 'sum']:
            # At this point method is not "bending"
            method: Union[Literal['centripetal'],
                          Literal['centrifugal'],
                          Literal['sum']]
            mmetrics.flow_centrality(x, mode=method)

    # Make copy, so that we don't screw things up
    x = x.copy()

    # Now get the node point with the highest flow centrality.
    cut = x.nodes[x.nodes.flow_centrality
                  == x.nodes.flow_centrality.max()].node_id.values

    # If more than one point we need to get one closest to the soma (root)
    if len(cut) > 1:
        cut = sorted(cut, key=lambda y: graph.dist_between(x.graph, y, x.root[0]))[0]
    else:
        cut = cut[0]

    if return_point:
        return cut

    # If cut node is a branch point, we will try cutting off main neurite
    if x.graph.degree(cut) > 2:
        # First make sure that there are no other branch points with flow
        # between this one and the soma
        path_to_root = nx.shortest_path(x.graph, cut, x.root[0])

        # Get flow centrality along the path
        flows = x.nodes.set_index('node_id', inplace=False).loc[path_to_root]

        # Subset to those that are branches (exclude mere synapses)
        flows = flows[flows.type == 'branch']

        # Find the first branch point from the soma with no flow (fillna is
        # important!)
        last_with_flow = np.where(flows.flow_centrality.fillna(0).values > 0)[0][-1]

        if method != 'bending':
            last_with_flow += 1

        to_cut = flows.iloc[last_with_flow].name

        # Cut off primary neurite
        rest, primary_neurite = graph.cut_neuron(x, to_cut)

        if method == 'bending':
            # The new cut node has to be a child of the original cut node
            cut = next(x.graph.predecessors(cut))

        # Change compartment and color
        primary_neurite.color = (0, 255, 0)  # type: ignore
        primary_neurite._register_attr('compartment', 'primary_neurite')  # type: ignore
    else:
        rest = x
        primary_neurite = None

    # Next, cut the rest into axon and dendrite
    a, b = graph.cut_neuron(rest, cut)

    # Figure out which one is which
    if split == 'prepost' or split == 'distance':
        a_inout = a.n_postsynapses / a.n_presynapses if a.n_presynapses else float('inf')
        b_inout = b.n_postsynapses / b.n_presynapses if b.n_presynapses else float('inf')
        if a_inout > b_inout:
            dendrite, axon = a, b
        else:
            dendrite, axon = b, a

    # Add compartment property
    axon._register_attr('compartment', 'axon')  # type: ignore
    dendrite._register_attr('compartment', 'dendrite')  # type: ignore

    # Change colors
    axon.color = (255, 0, 0)  # type: ignore
    dendrite.color = (0, 0, 255)  # type: ignore

    if primary_neurite:
        return core.NeuronList([primary_neurite, axon, dendrite])
    else:
        return core.NeuronList([axon, dendrite])


def stitch_neurons(*x: Union[Sequence[NeuronObject], 'core.NeuronList'],
                   method: Union[Literal['LEAFS'],
                                 Literal['ALL'],
                                 Literal['NONE']] = 'LEAFS',
                   master: Union[Literal['SOMA'],
                                 Literal['LARGEST'],
                                 Literal['FIRST']] = 'SOMA',
                   tn_to_stitch: Optional[Sequence[int]] = None,
                   suggest_only: bool = False,
                   max_dist: Optional[float] = None,
                   ) -> 'core.TreeNeuron':
    """Stitch multiple neurons together.

    Uses minimum spanning tree to determine a way to connect all fragments
    while minimizing length (eucledian distance) of the new edges. Nodes
    that have been stitched will get a "stitched" tag.

    Important
    ---------
    If duplicate node IDs are found across the fragments to stitch they will
    be remapped to new unique values!

    Parameters
    ----------
    x :                 NeuronList | list of TreeNeuron/List
                        Neurons to stitch (see examples).
    method :            'LEAFS' | 'ALL' | 'NONE', optional
                        Set stitching method:
                            (1) 'LEAFS': Only leaf (including root) nodes will
                                be allowed to make new edges.
                            (2) 'ALL': All nodes are considered.
                            (3) 'NONE': Node and connector tables will simply
                                be combined without generating any new edges.
                                The resulting neuron will have multiple roots.
    master :            'SOMA' | 'LARGEST' | 'FIRST', optional
                        Sets the master neuron:
                            (1) 'SOMA': The largest fragment with a soma
                                becomes the master neuron. If no neuron with
                                soma, will pick the largest (option 2).
                            (2) 'LARGEST': The largest fragment becomes the
                                master neuron.
                            (3) 'FIRST': The first fragment provided becomes
                                the master neuron.
    tn_to_stitch :      List of node IDs, optional
                        If provided, these nodes will be preferentially
                        used to stitch neurons together. Overrides methods
                        ``'ALL'`` or ``'LEAFS'``.
    suggest_only :      bool, optional
                        If True, will only return list of edges to add instead
                        of actually stitching the neuron.
    max_dist :          float,  optional
                        Max distance at which to stitch nodes. Setting this can
                        drastically speed up the process but can also lead to
                        failed stitching.

    Returns
    -------
    TreeNeuron
                        Stitched neuron.

    Examples
    --------
    Stitching neuronlist by simply combining data tables:

    >>> nl = navis.example_neurons(2)
    >>> stitched = navis.stitch_neurons(nl, method='NONE')

    Stitching fragmented neurons:
    >>> a = navis.example_neurons(1)
    >>> fragments = navis.cut_neuron(a, 100)
    >>> stitched = navis.stitch_neurons(frag, method='LEAFS')

    """
    method = str(method).upper()
    master = str(master).upper()

    if method not in ['LEAFS', 'ALL', 'NONE']:
        raise ValueError(f'Unknown method: "{method}"')

    if master not in ['SOMA', 'LARGEST', 'FIRST']:
        raise ValueError(f'Unknown master: "{master}"')

    # Compile list of individual neurons
    neurons = utils.unpack_neurons(x)

    # Use copies of the original neurons!
    nl = core.NeuronList(neurons).copy()

    if len(nl) < 2:
        logger.warning(f'Need at least 2 neurons to stitch, found {len(nl)}')
        return nl[0]

    # First find master
    if master == 'SOMA':
        has_soma = [n for n in nl if not isinstance(n.soma, type(None))]
        if len(has_soma) > 0:
            m = has_soma[0]
        else:
            m = sorted(nl.neurons,
                       key=lambda x: list(nl.cable_length),
                       reverse=True)[0]
    elif master == 'LARGEST':
        m = sorted(nl.neurons,
                   key=lambda x: list(nl.cable_length),
                   reverse=True)[0]
    else:
        # Simply pick the first neuron
        m = nl[0]

    # Check if we need to make any node IDs unique
    if nl.nodes.duplicated(subset='node_id').sum() > 0:
        seen_tn: Set[int] = set(m.nodes.node_id)
        for n in [n for n in nl if n != m]:
            this_tn = set(n.nodes.node_id)

            # Get duplicate node IDs
            non_unique = seen_tn & this_tn

            # Add this neuron's existing nodes to seen
            seen_tn = seen_tn | this_tn
            if non_unique:
                # Generate new, unique node IDs
                new_tn = np.arange(0, len(non_unique)) + max(seen_tn) + 1

                # Generate new map
                new_map = dict(zip(non_unique, new_tn))

                # Remap node IDs - if no new value, keep the old
                n.nodes.node_id = n.nodes.node_id.map(lambda x: new_map.get(x, x))

                if n.has_connectors:
                    n.connectors.node_id = n.connectors.node_id.map(lambda x: new_map.get(x, x))

                if hasattr(n, 'tags'):
                    n.tags = {new_map.get(k, k): v for k, v in n.tags.items()}  # type: ignore

                # Remap parent IDs
                new_map[None] = -1  # type: ignore
                n.nodes.parent_id = n.nodes.parent_id.map(lambda x: new_map.get(x, x)).astype(int)

                # Add new nodes to seen
                seen_tn = seen_tn | set(new_tn)

                # Make sure the graph is updated
                n._clear_temp_attr()

    # If method is none, we can just merge the data tables
    if method == 'NONE' or method is None:
        m.nodes = pd.concat([n.nodes for n in nl],  # type: ignore  # no stubs for concat
                            ignore_index=True)

        if any(nl.has_connectors):
            m.connectors = pd.concat([n.connectors for n in nl],  # type: ignore  # no stubs for concat
                                     ignore_index=True)

        m.tags = {}  # type: ignore  # TreeNeuron has no tags
        for n in nl:
            for k, v in getattr(n, 'tags', {}):
                m.tags[k] = m.tags.get(k, []) + list(utils.make_iterable(v))

        # Reset temporary attributes of our final neuron
        m._clear_temp_attr()

        return m

    # Fix potential problems with tn_to_stitch
    if not isinstance(tn_to_stitch, type(None)):
        if not isinstance(tn_to_stitch, (list, np.ndarray)):
            tn_to_stitch = [tn_to_stitch]

        # Make sure we're working with integers
        tn_to_stitch = [int(tn) for tn in tn_to_stitch]

    # Generate a list of potential new edges
    # Collect relevant nodes
    if not isinstance(tn_to_stitch, type(None)):
        tn = nl.nodes.loc[nl.nodes.node_id.isin(tn_to_stitch)]
    elif method == 'LEAFS':
        tn = nl.nodes.loc[nl.nodes['type'].isin(['end', 'root'])]
    else:
        tn = nl.nodes

    # Get pairwise distance between nodes
    # cdist
    d = scipy.spatial.distance.cdist(tn[['x', 'y', 'z']].values,
                                     tn[['x', 'y', 'z']].values,
                                     metric='euclidean')
    d = pd.DataFrame(d, index=tn.node_id.values, columns=tn.node_id.values)

    # Mask for the upper triangle
    mask = np.triu(np.ones(d.shape)).astype(np.bool)

    # Set lower triangle to NaN
    d = d.where(mask)

    # Stack
    new_edges = d.stack().reset_index(drop=False)
    new_edges.columns = ['source', 'target', 'weight']

    # Kick stuff that's too far away
    if max_dist:
        new_edges = new_edges[new_edges.weight <= max_dist]

    # Remove self edges
    new_edges = new_edges[new_edges.source != new_edges.target]

    # Remove edges that are within the same neuron
    id2ix = tn.set_index('node_id').neuron
    new_edges = new_edges[new_edges.source.map(id2ix) != new_edges.target.map(id2ix)]

    # Extract edges that need to be added to connect fragments
    if config.use_igraph and all([n.igraph for n in nl]):
        to_add = _mst_igraph(nl, new_edges)
    else:
        to_add = _mst_nx(nl, new_edges)

    if suggest_only:
        return to_add

    # Keep track of original master root
    master_root = m.root[0]

    # Generate one big neuron
    m.nodes = nl.nodes

    if any(nl.has_connectors):
        m.connectors = nl.connectors

    if any([hasattr(n, 'tags') for n in nl]):
        m.tags = {}  # type: ignore  # TreeNeuron has no tags
        for n in nl:
            m.tags.update(getattr(n, 'tags', {}))

    # Clear temporary attributes
    m._clear_temp_attr()

    for e in to_add:
        # Reroot to one of the nodes in the edge
        m.reroot(e[0], inplace=True)

        # Connect the nodes
        m.nodes.loc[m.nodes.node_id == e[0], 'parent_id'] = e[1]

        # Add edge to graphs
        if config.use_igraph and m.igraph:
            m.igraph.add_edge(m.igraph.vs.find(node_id=e[0]),
                              m.igraph.vs.find(node_id=e[1]),
                              **e[2])
        # We only really need to update this graph if we need it for reroot
        else:
            m.graph.add_edge(e[0], e[1], **e[2])

        # Add node tags
        m.tags = getattr(m, 'tags', {})  # type: ignore  # TreeNeuron has no tags
        m.tags['stitched'] = m.tags.get('stitched', []) + [e[0], e[1]]

        # We don't need this because reroot is taking care of this already
        # m._clear_temp_attr(exclude=['graph', 'igraph'])

    # Reroot to original root
    m.reroot(master_root, inplace=True)

    return m


def _mst_igraph(nl: 'core.NeuronList',
                new_edges: pd.DataFrame) -> List[List[int]]:
    """Compute edges necessary to connect a fragmented neuron using igraph."""
    # Generate a union of all graphs
    g = nl[0].igraph.disjoint_union(nl[1:].igraph)

    # We have to manually set the node IDs again
    nids = np.concatenate([n.igraph.vs['node_id'] for n in nl])
    g.vs['node_id'] = nids

    # Set existing edges to zero weight to make sure they have priority when
    # calculating the minimum spanning tree
    g.es['weight'] = 0

    # If two nodes occupy the same position (e.g. after if fragments are the
    # result of cutting), they will have a distance of 0. Hence, we won't be
    # able to simply filter by distance
    g.es['new'] = False

    # Convert node IDs in new_edges to vertex IDs and add to graph
    name2ix = dict(zip(g.vs['node_id'], range(len(g.vs))))
    new_edges['source_ix'] = new_edges.source.map(name2ix)
    new_edges['target_ix'] = new_edges.target.map(name2ix)

    # Add new edges
    g.add_edges(new_edges[['source_ix', 'target_ix']].values.tolist())

    # Add edge weight to new edges
    g.es[-new_edges.shape[0]:]['weight'] = new_edges.weight.values

    # Keep track of new edges
    g.es[-new_edges.shape[0]:]['new'] = True

    # Compute the minimum spanning tree
    mst = g.spanning_tree(weights='weight')

    # Extract the new edges
    to_add = mst.es.select(new=True)

    # Convert to node IDs
    to_add = [(g.vs[e.source]['node_id'],
               g.vs[e.target]['node_id'],
               {'weight': e['weight']})
              for e in to_add]

    return to_add


def _mst_nx(nl: 'core.NeuronList',
            new_edges: pd.DataFrame) -> List[List[int]]:
    """Compute edges necessary to connect a fragmented neuron using networkX."""
    # Generate a union of all graphs
    g = nx.union_all([n.graph for n in nl]).to_undirected()

    # Set existing edges to zero weight to make sure they have priority when
    # calculating the minimum spanning tree
    nx.set_edge_attributes(g, 0, 'weight')

    # If two nodes occupy the same position (e.g. after if fragments are the
    # result of cutting), they will have a distance of 0. Hence, we won't be
    # able to simply filter by distance
    nx.set_edge_attributes(g, False, 'new')

    # Convert new edges in the right format
    edges_nx = [(r.source, r.target, {'weight': r.weight, 'new': True})
                for r in new_edges.itertuples()]

    # Add edges to union graph
    g.add_edges_from(edges_nx)

    # Get minimum spanning tree
    edges = nx.minimum_spanning_edges(g)

    # Edges that need adding are those that were newly added
    to_add = [e for e in edges if e[2]['new']]

    return to_add


def average_neurons(x: 'core.NeuronList',
                    limit: int = 10,
                    base_neuron: Optional[Union[int, 'core.TreeNeuron']] = None
                    ) -> 'core.TreeNeuron':
    """ Computes an average from a list of neurons.

    This is a very simple implementation which may give odd results if used
    on complex neurons. Works fine on e.g. backbones or tracts.

    Parameters
    ----------
    x :             NeuronList
                    Neurons to be averaged.
    limit :         int, optional
                    Max distance for nearest neighbour search.
    base_neuron :   neuron id | TreeNeuron, optional
                    Neuron to use as template for averaging. If not provided,
                    the first neuron in the list is used as template!

    Returns
    -------
    TreeNeuron

    Examples
    --------
    >>> # Get a bunch of neurons
    >>> da2 = navis.example_neurons()
    >>> # Prune down to longest neurite
    >>> da2.reroot(da2.soma)
    >>> da2_pr = da2.prune_by_longest_neurite(inplace=False)
    >>> # Make average
    >>> da2_avg = navis.average_neurons(da2_pr, limit=10e3)
    >>> # Plot
    >>> da2.plot3d()
    >>> da2_avg.plot3d()

    """

    if not isinstance(x, core.NeuronList):
        raise TypeError(f'Need NeuronList, got "{type(x)}"')

    if len(x) < 2:
        raise ValueError('Need at least 2 neurons to average!')

    # Generate KDTrees for each neuron
    for n in x:
        n.tree = graph.neuron2KDTree(n, tree_type='c', data='nodes')  # type: ignore  # TreeNeuron has no tree

    # Set base for average: we will use this neurons nodes to query
    # the KDTrees
    if isinstance(base_neuron, core.TreeNeuron):
        bn = base_neuron.copy()
    elif isinstance(base_neuron, int):
        bn = x[base_neuron].copy()
    elif isinstance(base_neuron, type(None)):
        bn = x[0].copy()
    else:
        raise ValueError('Unable to interpret base_neuron of '
                         'type "{0}"'.format(type(base_neuron)))

    base_nodes = bn.nodes[['x', 'y', 'z']].values
    other_neurons = x[1:]

    # Make sure these stay 2-dimensional arrays -> will add a colum for each
    # "other" neuron
    base_x = base_nodes[:, 0:1]
    base_y = base_nodes[:, 1:2]
    base_z = base_nodes[:, 2:3]

    # For each "other" neuron, collect nearest neighbour coordinates
    for n in other_neurons:
        nn_dist, nn_ix = n.tree.query(base_nodes,
                                      k=1,
                                      distance_upper_bound=limit)

        # Translate indices into coordinates
        # First, make empty array
        this_coords = np.zeros((len(nn_dist), 3))
        # Set coords without a nearest neighbour within distances to "None"
        this_coords[nn_dist == float('inf')] = None
        # Fill in coords of nearest neighbours
        this_coords[nn_dist != float(
            'inf')] = n.tree.data[nn_ix[nn_dist != float('inf')]]
        # Add coords to base coords
        base_x = np.append(base_x, this_coords[:, 0:1], axis=1)
        base_y = np.append(base_y, this_coords[:, 1:2], axis=1)
        base_z = np.append(base_z, this_coords[:, 2:3], axis=1)

    # Calculate means
    mean_x = np.mean(base_x, axis=1)
    mean_y = np.mean(base_y, axis=1)
    mean_z = np.mean(base_z, axis=1)

    # If any of the base coords has NO nearest neighbour within limit
    # whatsoever, the average of that row will be "NaN" -> in this case we
    # will fall back to the base coordinate
    mean_x[np.isnan(mean_x)] = base_nodes[np.isnan(mean_x), 0]
    mean_y[np.isnan(mean_y)] = base_nodes[np.isnan(mean_y), 1]
    mean_z[np.isnan(mean_z)] = base_nodes[np.isnan(mean_z), 2]

    # Change coordinates accordingly
    bn.nodes.loc[:, 'x'] = mean_x
    bn.nodes.loc[:, 'y'] = mean_y
    bn.nodes.loc[:, 'z'] = mean_z

    return bn


def despike_neuron(x: NeuronObject,
                   sigma: int = 5,
                   max_spike_length: int = 1,
                   inplace: bool = False,
                   reverse: bool = False) -> Optional[NeuronObject]:
    """ Removes spikes in neuron traces (e.g. from jumps in image data).

    For each node A, the euclidean distance to its next successor (parent)
    B and that node's successor is computed. If
    :math:`\\frac{dist(A,B)}{dist(A,C)}>sigma`, node B is considered a spike
    and realigned between A and C.

    Parameters
    ----------
    x :                 TreeNeuron | NeuronList
                        Neuron(s) to be processed.
    sigma :             float | int, optional
                        Threshold for spike detection. Smaller sigma = more
                        aggressive spike detection.
    max_spike_length :  int, optional
                        Determines how long (# of nodes) a spike can be.
    inplace :           bool, optional
                        If False, a copy of the neuron is returned.
    reverse :           bool, optional
                        If True, will **also** walk the segments from proximal
                        to distal. Use this to catch spikes on e.g. terminal
                        nodes.

    Returns
    -------
    TreeNeuron/List
                Despiked neuron(s). Only if ``inplace=False``.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> despiked = navis.despike_neuron(n)

    """
    # TODO:
    # - flattening all segments first before Spike detection should speed up
    #   quite a lot
    # -> as intermediate step: assign all new positions at once

    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()

        for n in config.tqdm(x, desc='Despiking', disable=config.pbar_hide,
                             leave=config.pbar_leave):
            despike_neuron(n, sigma=sigma, inplace=True)

        if not inplace:
            return x
        return None
    elif not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Can only process TreeNeuron or NeuronList, not {type(x)}')

    if not inplace:
        x = x.copy()

    # Index nodes table by node ID
    this_nodes = x.nodes.set_index('node_id', inplace=False)

    segs_to_walk = x.segments

    if reverse:
        segs_to_walk += x.segments[::-1]

    # For each spike length do -> do this in reverse to correct the long
    # spikes first
    for l in list(range(1, max_spike_length + 1))[::-1]:
        # Go over all segments
        for seg in x.segments:
            # Get nodes A, B and C of this segment
            this_A = this_nodes.loc[seg[:-l - 1]]
            this_B = this_nodes.loc[seg[l:-1]]
            this_C = this_nodes.loc[seg[l + 1:]]

            # Get coordinates
            A = this_A[['x', 'y', 'z']].values
            B = this_B[['x', 'y', 'z']].values
            C = this_C[['x', 'y', 'z']].values

            # Calculate euclidian distances A->B and A->C
            dist_AB = np.linalg.norm(A - B, axis=1)
            dist_AC = np.linalg.norm(A - C, axis=1)

            # Get the spikes
            spikes_ix = np.where((dist_AB / dist_AC) > sigma)[0]
            spikes = this_B.iloc[spikes_ix]

            if not spikes.empty:
                # Interpolate new position(s) between A and C
                new_positions = A[spikes_ix] + (C[spikes_ix] - A[spikes_ix]) / 2

                this_nodes.loc[spikes.index, ['x', 'y', 'z']] = new_positions

    # Reassign node table
    x.nodes = this_nodes.reset_index(drop=False, inplace=False)

    # The weights in the graph have changed, we need to update that
    x._clear_temp_attr(exclude=['segments', 'small_segments',
                                'classify_nodes'])

    if not inplace:
        return x
    else:
        return None


def guess_radius(x: NeuronObject,
                 method: str = 'linear',
                 limit: Optional[int] = None,
                 smooth: bool = True,
                 inplace: bool = False) -> Optional[NeuronObject]:
    """ Tries guessing radii for all nodes.

    Uses distance between connectors and nodes and interpolate for all
    nodes. Fills in ``radius`` column in node table.

    Parameters
    ----------
    x :             TreeNeuron | NeuronList
                    Neuron(s) to be processed.
    method :        str, optional
                    Method to be used to interpolate unknown radii. See
                    ``pandas.DataFrame.interpolate`` for details.
    limit :         int, optional
                    Maximum number of consecutive missing radii to fill.
                    Must be greater than 0.
    smooth :        bool | int, optional
                    If True, will smooth radii after interpolation using a
                    rolling window. If ``int``, will use to define size of
                    window.
    inplace :       bool, optional
                    If False, will use and return copy of original neuron(s).

    Returns
    -------
    TreeNeuron/List
                    If ``inplace=False``.

    """
    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()

        for n in config.tqdm(x, desc='Guessing', disable=config.pbar_hide,
                             leave=config.pbar_leave):
            guess_radius(n, method=method, limit=limit, smooth=smooth,
                         inplace=True)

        if not inplace:
            return x
        return None

    elif not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Can only process TreeNeuron or NeuronList, not {type(x)}')

    if not hasattr(x, 'connectors') or x.connectors.empty:
        raise ValueError('Neuron must have connectors!')

    if not inplace:
        x = x.copy()

    # Set default rolling window size
    if isinstance(smooth, bool) and smooth:
        smooth = 5

    # We will be using the index as distance to interpolate. For this we have
    # to change method 'linear' to 'index'
    method = 'index' if method == 'linear' else method

    # Collect connectors and calc distances
    cn = x.connectors.copy()

    # Prepare nodes (add parent_dist for later, set index)
    mmetrics.parent_dist(x, root_dist=0)
    nodes = x.nodes.set_index('node_id', inplace=False)

    # For each connector (pre and post), get the X/Y distance to its node
    cn_locs = cn[['x', 'y']].values
    tn_locs = nodes.loc[cn.node_id.values,
                        ['x', 'y']].values
    dist = np.sqrt(np.sum((tn_locs - cn_locs) ** 2, axis=1).astype(int))
    cn['dist'] = dist

    # Get max distance per node (in case of multiple connectors per
    # node)
    cn_grouped = cn.groupby('node_id').max()

    # Set undefined radii to None
    nodes.loc[nodes.radius <= 0, 'radius'] = None

    # Assign radii to nodes
    nodes.loc[cn_grouped.index, 'radius'] = cn_grouped.dist.values

    # Go over each segment and interpolate radii
    for s in config.tqdm(x.segments, desc='Interp.', disable=config.pbar_hide,
                         leave=config.pbar_leave):

        # Get this segments radii and parent dist
        this_radii = nodes.loc[s, ['radius', 'parent_dist']]
        this_radii['parent_dist_cum'] = this_radii.parent_dist.cumsum()

        # Set cumulative distance as index and drop parent_dist
        this_radii = this_radii.set_index('parent_dist_cum',
                                          drop=True).drop('parent_dist',
                                                          axis=1)

        # Interpolate missing radii
        interp = this_radii.interpolate(method=method, limit_direction='both',
                                        limit=limit)

        if smooth:
            interp = interp.rolling(smooth,
                                    min_periods=1).max()

        nodes.loc[s, 'radius'] = interp.values

    # Set non-interpolated radii back to -1
    nodes.loc[nodes.radius.isnull(), 'radius'] = -1

    # Reassign nodes
    x.nodes = nodes.reset_index(drop=False, inplace=False)

    if not inplace:
        return x
    else:
        return None


def smooth_neuron(x: NeuronObject,
                  window: int = 5,
                  inplace: bool = False) -> Optional[NeuronObject]:
    """ Smooth neuron using rolling windows.

    Parameters
    ----------
    x :             TreeNeuron | NeuronList
                    Neuron(s) to be processed.
    window :        int, optional
                    Size (n observations) of the rolling window in number of
                    nodes.
    inplace :       bool, optional
                    If False, will use and return copy of original neuron(s).

    Returns
    -------
    TreeNeuron/List
                    Smoothed neuron(s). If ``inplace=False``.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> smoothed = navis.smooth_neuron(n, window=10)

    """

    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()

        for n in config.tqdm(x, desc='Smoothing', disable=config.pbar_hide,
                             leave=config.pbar_leave):
            smooth_neuron(n, window=window, inplace=True)

        if not inplace:
            return x
        else:
            return None

    elif not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Can only process TreeNeuron or NeuronList, not {type(x)}')

    if not inplace:
        x = x.copy()

    # Prepare nodes (add parent_dist for later, set index)
    mmetrics.parent_dist(x, root_dist=0)
    nodes = x.nodes.set_index('node_id', inplace=False).copy()

    # Go over each segment and smooth
    for s in config.tqdm(x.segments, desc='Smoothing',
                         disable=config.pbar_hide,
                         leave=config.pbar_leave):

        # Get this segment's parent distances and get cumsum
        this_co = nodes.loc[s, ['x', 'y', 'z']]

        interp = this_co.rolling(window, min_periods=1).mean()

        nodes.loc[s, ['x', 'y', 'z']] = interp.values

    # Reassign nodes
    x.nodes = nodes.reset_index(drop=False, inplace=False)

    x._clear_temp_attr()

    if not inplace:
        return x
    else:
        return None


def break_fragments(x: 'core.TreeNeuron') -> 'core.NeuronList':
    """Break neuron into continuous fragments.

    Neurons can consists of several disconnected fragments. This function
    turn these fragments into separate neurons.

    Parameters
    ----------
    x :         TreeNeuron
                Fragmented neuron.

    Returns
    -------
    NeuronList

    See Also
    --------
    :func:`navis.heal_fragmented_neuron`
                Use to heal fragmentation instead of breaking it up.


    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> # Disconnect parts of the neuron
    >>> n.nodes.loc[100, 'parent_id'] = -1
    >>> # Break into fragments
    >>> frags = navis.break_fragments(n)
    >>> len(frags)
    2

    """
    if isinstance(x, core.NeuronList) and len(x) == 1:
        x = x[0]

    if not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Expected Neuron/List, got "{type(x)}"')

    # Don't do anything if not actually fragmented
    if x.n_skeletons > 1:
        # Get connected components
        comp = graph._connected_components(x)
        # Sort so that the first component is the largest
        comp = sorted(comp, key=len, reverse=True)

        return core.NeuronList([graph.subset_neuron(x,
                                                    list(ss),
                                                    inplace=False) for ss in comp])
    else:
        return core.NeuronList(x.copy())


def heal_fragmented_neuron(x: 'core.NeuronList',
                           min_size: int = 0,
                           method: Union[Literal['LEAFS'],
                                         Literal['ALL']] = 'LEAFS',
                           max_dist: Optional[float] = None,
                           inplace: bool = False) -> Optional[NeuronObject]:
    """ Heal fragmented neuron(s).

    Tries to heal a fragmented neuron (i.e. a neuron with multiple roots)
    using a minimum spanning tree.

    Parameters
    ----------
    x :         TreeNeuron/List
                Fragmented neuron(s).
    min_size :  int, optional
                Minimum size in nodes for fragments to be reattached.
    method :    'LEAFS' | 'ALL', optional
                Method used to heal fragments:
                        (1) 'LEAFS': Only leaf (including root) nodes will
                            be used to heal gaps.
                        (2) 'ALL': All nodes can be used to reconnect
                            fragments.
    max_dist :  float, optional
                Max distance at which to merge fragments. Setting this to a
                reasonable value can dramatically speed things up. If too low,
                this can lead to failed healing.
    inplace :   bool, optional
                If False, will perform healing on and return a copy.

    Returns
    -------
    None
                If ``inplace=True``
    CatmaidNeuron/List
                If ``inplace=False``


    See Also
    --------
    :func:`navis.stitch_neurons`
                Function used by ``heal_fragmented_neuron`` to stitch
                fragments.
    :func:`navis.break_fragments`
                Use to break a fragmented neuron into disconnected pieces.


    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> # Disconnect parts of the neuron
    >>> n.nodes.loc[100, 'parent_id'] = -1
    >>> len(n.root)
    2
    >>> # Heal neuron
    >>> healed = navis.heal_fragmented_neuron(n)
    >>> len(healed.root)
    1
    """

    method = str(method).upper()

    if method not in ['LEAFS', 'ALL']:
        raise ValueError(f'Unknown method "{method}"')

    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()
        healed = [heal_fragmented_neuron(n,
                                         min_size=min_size,
                                         method=method,
                                         max_dist=max_dist,
                                         inplace=True)
                  for n in config.tqdm(x,
                                       desc='Healing',
                                       disable=config.pbar_hide,
                                       leave=config.pbar_leave)]
        if not inplace:
            return x
        else:
            return None

    if not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Expected CatmaidNeuron/List, got "{type(x)}"')

    # Only process if actually fragmented
    if x.n_skeletons > 1:
        frags = break_fragments(x)
        healed = stitch_neurons(*[f for f in frags if f.n_nodes > min_size],
                                max_dist=max_dist,
                                method=method)
        if not inplace:
            return healed
        else:
            x.nodes = healed.nodes  # update nodes
            x.tags = healed.tags  # update tags
            x._clear_temp_attr()
            return None
    else:
        if not inplace:
            return x
        else:
            return None



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2017 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along

""" Module contains functions to plot neurons in 2D and 3D.
"""

import random
import math

import numpy as np

from typing import Tuple, Optional, List, Dict

from .. import config, core

__all__ = ['tn_pairs_to_coords', 'segments_to_coords', 'fibonacci_sphere']

logger = config.logger


def tn_pairs_to_coords(x: core.TreeNeuron,
                       modifier: Optional[Tuple[float,
                                                float,
                                                float]] = (1, 1, 1)
                       ) -> np.ndarray:
    """Returns pairs of treenode -> parent node coordinates.

    Parameters
    ----------
    x :         pandas DataFrame | TreeNeuron
                Must contain the nodes.
    modifier :  ints, optional
                Use to modify/invert x/y/z axes.

    Returns
    -------
    coords :    np.array
                ``[[[x1, y1, z1], [x2, y2, z2]], [[x3, y3, y4], [x4, y4, z4]]]``

    """

    if not isinstance(modifier, np.ndarray):
        modifier = np.array(modifier)

    nodes = x.nodes[x.nodes.parent_id >= 0]

    tn_co = nodes.loc[:, ['x', 'y', 'z']].values
    parent_co = x.nodes.set_index('node_id',
                                  inplace=False).loc[nodes.parent_id.values,
                                                     ['x', 'y', 'z']].values

    tn_co *= modifier
    parent_co *= modifier

    coords = np.append(tn_co, parent_co, axis=1)

    return coords.reshape((coords.shape[0], 2, 3))


def segments_to_coords(x: core.TreeNeuron,
                       segments: List[List[int]],
                       modifier: Optional[Tuple[float,
                                                float,
                                                float]] = (1, 1, 1)
                       ) -> List[np.ndarray]:
    """ Turns lists of node IDs into coordinates.

    Parameters
    ----------
    x :         pandas DataFrame | TreeNeuron
                Must contain the nodes
    segments :  list of treenode IDs
    modifier :  ints, optional
                Use to modify/invert x/y/z axes.

    Returns
    -------
    coords :    list of tuples
                [(x, y, z), (x, y, z), ... ]

    """

    if not isinstance(modifier, np.ndarray):
        modifier = np.array(modifier)

    locs: Dict[int, Tuple[float, float, float]]
    locs = {r.node_id: (r.x, r.y, r.z) for r in x.nodes.itertuples()}  # type: ignore

    coords = ([np.array([locs[tn] for tn in s]) * modifier for s in segments])

    return coords


def fibonacci_sphere(samples: int = 1,
                     randomize: bool = True) -> list:
    """ Calculates points on a sphere
    """
    rnd = 1.
    if randomize:
        rnd = random.random() * samples

    points = []
    offset = 2. / samples
    increment = math.pi * (3. - math.sqrt(5.))

    for i in range(samples):
        y = ((i * offset) - 1) + (offset / 2)
        r = math.sqrt(1 - pow(y, 2))

        phi = ((i + rnd) % samples) * increment

        x = math.cos(phi) * r
        z = math.sin(phi) * r

        points.append([x, y, z])

    return points



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2017 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along

import uuid

import pandas as pd
import numpy as np

from typing import Tuple, Iterable, List, Union, Any, Optional, Sequence

from .. import config, core

from .iterables import *

# Set up logging
logger = config.logger


def eval_conditions(x) -> Tuple[List[bool], List[bool]]:
    """ Splits list of strings into positive (no ~) and negative (~) conditions
    """

    x = make_iterable(x, force_type=str)

    return [i for i in x if not i.startswith('~')], [i[1:] for i in x if i.startswith('~')]


def eval_id(x: Union[uuid.UUID, str, 'core.NeuronObject', pd.DataFrame],
              warn_duplicates: bool = True) -> List[uuid.UUID]:
    """Evaluate neurons' ID(s).

    Parameters
    ----------
    x :                str | uuid.UUID | TreeNeuron | NeuronList | DataFrame
                       For Neuron/List or pandas.DataFrames/Series will
                       look for ``id`` attribute/column.
    warn_duplicates :  bool, optional
                       If True, will warn if duplicate IDs are found.
                       Only applies to NeuronLists.

    Returns
    -------
    list
                    List containing IDs.

    """
    if isinstance(x, (uuid.UUID, str, np.str, int, np.int64, np.int)):
        return [x]
    elif isinstance(x, (list, np.ndarray, set)):
        uu: List[uuid.UUID] = []
        for e in x:
            temp = eval_id(e, warn_duplicates=warn_duplicates)
            if isinstance(temp, (list, np.ndarray)):
                uu += temp
            else:
                uu.append(temp)  # type: ignore
        return sorted(set(uu), key=uu.index)
    elif isinstance(x, core.TreeNeuron):
        return [x.id]
    elif isinstance(x, core.NeuronList):
        if len(x.id) != len(set(x.id)) and warn_duplicates:
            logger.warning('Duplicate IDs found in NeuronList.'
                           'The function you are using might not respect '
                           'fragments of the same neuron. For explanation see '
                           'http://navis.readthedocs.io/en/latest/source/conn'
                           'ectivity_analysis.html.')
        return list(x.id)
    elif isinstance(x, pd.DataFrame):
        if 'id' not in x.columns:
            raise ValueError('Expect "id" column in pandas DataFrames')
        return x.id.tolist()
    elif isinstance(x, pd.Series):
        if x.name == 'id':
            return x.tolist()
        elif 'id' in x:
            return [x.id]
        else:
            raise ValueError(f'Unable to extract ID from pandas series {x}')
    elif isinstance(x, type(None)):
        return None
    else:
        msg = f'Unable to extract ID(s) from data of type "{type(x)}"'
        logger.error(msg)
        raise TypeError(msg)


def eval_neurons(x: Any,
                 warn_duplicates: bool = True,
                 raise_other: bool = True) -> Optional[List['core.TreeNeuron']]:
    """ Evaluate neurons.

    Parameters
    ----------
    x
                       Data to be checked for neurons.
    warn_duplicates :  bool, optional
                       If True, will warn if duplicate neurons are found.
                       Only applies to NeuronLists.
    raise_other :      bool, optional
                       If True, will raise error if non- neurons are found.

    Returns
    -------
    list
                    List containing neurons.
    None
                    If no neurons found.

    """

    if isinstance(x, core.TreeNeuron):
        return [x]
    elif isinstance(x, (list, np.ndarray, set)):
        neurons: List['core.TreeNeuron'] = []
        for e in x:
            temp = eval_neurons(e, warn_duplicates=warn_duplicates,
                                raise_other=raise_other)
            if isinstance(temp, (list, np.ndarray)):
                neurons += temp
            elif temp:
                neurons.append(temp)
        return sorted(set(neurons), key=neurons.index)
    elif isinstance(x, core.TreeNeuron):
        return [x]
    elif isinstance(x, core.NeuronList):
        if len(x.id) != len(set(x.id)) and warn_duplicates:
            logger.warning('Duplicate IDs found in NeuronList.'
                           'The function you are using might not respect '
                           'fragments of the same neuron. For explanation see '
                           'http://navis.readthedocs.io/en/latest/source/conn'
                           'ectivity_analysis.html.')
        return x.neurons
    elif isinstance(x, type(None)):
        return None
    elif raise_other:
        msg = f'Unable to extract neurons from data of type "{type(x)}"'
        logger.error(msg)
        raise TypeError(msg)
    return None


def eval_node_ids(x: Union[int, str,
                           Sequence[Union[str, int]],
                           'core.NeuronObject',
                           pd.DataFrame]
                  ) -> List[int]:
    """ Extract node IDs from data.

    Parameters
    ----------
    x :             int | str | TreeNeuron | NeuronList | DataFrame
                    Your options are either::
                    1. int or list of ints will be assumed to be node IDs
                    2. str or list of str will be checked if convertible to int
                    3. For TreeNeuron/List or pandas.DataFrames will try
                       to extract node IDs

    Returns
    -------
    list
                    List containing node IDs (integer)

    """

    if isinstance(x, (int, np.int64, np.int32, np.int)):
        return [x]
    elif isinstance(x, (str, np.str)):
        try:
            return [int(x)]
        except BaseException:
            raise TypeError(f'Unable to extract node ID from string "{x}"')
    elif isinstance(x, (set, list, np.ndarray)):
        # Check non-integer entries
        ids: List[int] = []
        for e in x:
            temp = eval_node_ids(e)
            if isinstance(temp, (list, np.ndarray)):
                ids += temp
            else:
                ids.append(temp)  # type: ignore
        # Preserving the order after making a set is super costly
        # return sorted(set(ids), key=ids.index)
        return list(set(ids))
    elif isinstance(x, core.TreeNeuron):
        return x.nodes.node_id.astype(int).tolist()
    elif isinstance(x, core.NeuronList):
        to_return: List[int] = []
        for n in x:
            to_return += n.nodes.node_id.astype(int).tolist()
        return to_return
    elif isinstance(x, (pd.DataFrame, pd.Series)):
        to_return = []
        if 'node_id' in x:
            to_return += x.node_id.astype(int).tolist()
        return to_return
    else:
        raise TypeError(f'Unable to extract node IDs from type {type(x)}')
